#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import unittest
import pero


class TestCase(unittest.TestCase):
    """Test case for Color class."""
    
    
    def test_valid_constructor(self):
        """Tests whether constructor works correctly."""
        
        # test individual channels
        color = pero.Color(100, 150, 200, 10)
        
        self.assertEqual(color.red, 100)
        self.assertEqual(color.green, 150)
        self.assertEqual(color.blue, 200)
        self.assertEqual(color.alpha, 10)
        
        color = pero.Color(100, 150, 200)
        
        self.assertEqual(color.red, 100)
        self.assertEqual(color.green, 150)
        self.assertEqual(color.blue, 200)
        self.assertEqual(color.alpha, 255)
        
        # test tuple
        color = pero.Color((100, 150, 200, 10))
        
        self.assertEqual(color.red, 100)
        self.assertEqual(color.green, 150)
        self.assertEqual(color.blue, 200)
        self.assertEqual(color.alpha, 10)
        
        color = pero.Color((100, 150, 200))
        
        self.assertEqual(color.red, 100)
        self.assertEqual(color.green, 150)
        self.assertEqual(color.blue, 200)
        self.assertEqual(color.alpha, 255)
        
        # test hex
        color = pero.Color("#6496c80a")
        
        self.assertEqual(color.red, 100)
        self.assertEqual(color.green, 150)
        self.assertEqual(color.blue, 200)
        self.assertEqual(color.alpha, 10)
        
        color = pero.Color("#6496c8")
        
        self.assertEqual(color.red, 100)
        self.assertEqual(color.green, 150)
        self.assertEqual(color.blue, 200)
        self.assertEqual(color.alpha, 255)
    
    
    def test_invalid_constructor(self):
        """Tests whether constructor works correctly."""
        
        # test incorrect type
        with self.assertRaises(ValueError):
            color = pero.Color("red", 150, 200, 10)
        with self.assertRaises(ValueError):
            color = pero.Color(100, "green", 200, 10)
        with self.assertRaises(ValueError):
            color = pero.Color(100, 150, "blue", 100)
        with self.assertRaises(ValueError):
            color = pero.Color(100, 150, 200, "alpha")
        
        # test incorrect values
        with self.assertRaises(ValueError):
            color = pero.Color(300, 150, 200, 10)
        with self.assertRaises(ValueError):
            color = pero.Color(100, 300, 200, 10)
        with self.assertRaises(ValueError):
            color = pero.Color(100, 150, 300, 100)
        with self.assertRaises(ValueError):
            color = pero.Color(100, 150, 200, 300)
        with self.assertRaises(ValueError):
            color = pero.Color(-1, 150, 200, 10)
        with self.assertRaises(ValueError):
            color = pero.Color(100, -1, 200, 10)
        with self.assertRaises(ValueError):
            color = pero.Color(100, 150, -1, 100)
        with self.assertRaises(ValueError):
            color = pero.Color(100, 150, 200, -1)
    
    
    def test_equals(self):
        """Tests whether equality comparer works correctly."""
        
        # test full equal
        c1 = pero.Color(100, 150, 200, 255, name="Name")
        c2 = pero.Color(100, 150, 200, 255, name="Name")
        
        self.assertEqual(c1, c2)
        self.assertEqual(c1, c2.rgba)
        
        # test equal channels
        c1 = pero.Color(100, 150, 200, 255, name="Name1")
        c2 = pero.Color(100, 150, 200, 255, name="Name2")
        
        self.assertEqual(c1, c2)
        self.assertEqual(c1, c2.rgba)
        
        # test non-equal channels
        c1 = pero.Color(101, 150, 200, 255)
        c2 = pero.Color(100, 150, 200, 255)
        
        self.assertNotEqual(c1, c2)
        self.assertNotEqual(c1, c2.rgba)
        
        c1 = pero.Color(100, 151, 200, 255)
        c2 = pero.Color(100, 150, 200, 255)
        
        self.assertNotEqual(c1, c2)
        self.assertNotEqual(c1, c2.rgba)
        
        c1 = pero.Color(100, 150, 201, 255)
        c2 = pero.Color(100, 150, 200, 255)
        
        self.assertNotEqual(c1, c2)
        self.assertNotEqual(c1, c2.rgba)
        
        c1 = pero.Color(100, 150, 201, 251)
        c2 = pero.Color(100, 150, 200, 255)
        
        self.assertNotEqual(c1, c2)
        self.assertNotEqual(c1, c2.rgba)
    
    
    def test_rgba(self):
        """Tests whether RGBA channels are returned correctly."""
        
        color = pero.Color(100, 150, 200, 250)
        self.assertEqual(color.rgba, (100, 150, 200, 250))
        
        color = pero.Color(100, 150, 200, 250)
        self.assertEqual(color.rgba_r, (100/255, 150/255, 200/255, 250/255))
        
        color = pero.Color(100, 150, 200, 250)
        self.assertEqual(color.rgb, (100, 150, 200))
        
        color = pero.Color(100, 150, 200, 250)
        self.assertEqual(color.rgb_r, (100/255, 150/255, 200/255))
    
    
    def test_hex(self):
        """Tests whether RGBA hex is returned correctly."""
        
        color = pero.Color(0, 150, 200, 250)
        self.assertEqual(color.hex, "#0096c8fa")
    
    
    def test_lighter(self):
        """Tests whether making lighter color works correctly."""
        
        color = pero.Color(50, 100, 180, 200)
        
        lighter = color.lighter(0)
        self.assertEqual(lighter.rgba, (50, 100, 180, 200))
        
        lighter = color.lighter(0.5)
        self.assertEqual(lighter.rgba, (153, 178, 218, 200))
        
        lighter = color.lighter(1)
        self.assertEqual(lighter.rgba, (255, 255, 255, 200))
        
        # test incorrect value
        with self.assertRaises(ValueError):
            trans = color.lighter(-1)
        
        with self.assertRaises(ValueError):
            trans = color.lighter(2)
    
    
    def test_darker(self):
        """Tests whether making darker color works correctly."""
        
        color = pero.Color(152, 177, 217, 200)
        
        darker = color.darker(0)
        self.assertEqual(darker.rgba, (152, 177, 217, 200))
        
        darker = color.darker(0.5)
        self.assertEqual(darker.rgba, (76, 89, 109, 200))
        
        darker = color.darker(1)
        self.assertEqual(darker.rgba, (0, 0, 0, 200))
        
        # test incorrect value
        with self.assertRaises(ValueError):
            trans = color.darker(-1)
        
        with self.assertRaises(ValueError):
            trans = color.darker(2)
    
    
    def test_opaque(self):
        """Tests whether making opaque color works correctly."""
        
        color = pero.Color(152, 177, 217, 200)
        
        trans = color.opaque(0)
        self.assertEqual(trans.rgba, (152, 177, 217, 0))
        
        trans = color.opaque(1)
        self.assertEqual(trans.rgba, (152, 177, 217, 255))
        
        trans = color.opaque(0.5)
        self.assertEqual(trans.rgba, (152, 177, 217, 128))
        
        # test incorrect value
        with self.assertRaises(ValueError):
            trans = color.opaque(-1)
        
        with self.assertRaises(ValueError):
            trans = color.opaque(2)
    
    
    def test_trans(self):
        """Tests whether making transparent color works correctly."""
        
        color = pero.Color(152, 177, 217, 200)
        
        trans = color.trans(0)
        self.assertEqual(trans.rgba, (152, 177, 217, 255))
        
        trans = color.trans(1)
        self.assertEqual(trans.rgba, (152, 177, 217, 0))
        
        trans = color.trans(0.5)
        self.assertEqual(trans.rgba, (152, 177, 217, 128))
        
        # test incorrect value
        with self.assertRaises(ValueError):
            trans = color.trans(-1)
        
        with self.assertRaises(ValueError):
            trans = color.trans(2)
    
    
    def test_interpolate(self):
        """Tests whether interpolation works correctly."""
        
        c1 = pero.Color(50, 100, 150, 200)
        c2 = pero.Color(100, 150, 200, 250)
        
        inter = pero.Color.interpolate(c1, c2, 0.5)
        self.assertEqual(inter.rgba, (75, 125, 175, 225))
        
        inter = pero.Color.interpolate(c1, c2, -0.5)
        self.assertEqual(inter.rgba, (25, 75, 125, 175))
        
        inter = pero.Color.interpolate(c1, c2, -1000)
        self.assertEqual(inter.rgba, (0, 0, 0, 0))
        
        inter = pero.Color.interpolate(c1, c2, 1.5)
        self.assertEqual(inter.rgba, (125, 175, 225, 255))
        
        inter = pero.Color.interpolate(c1, c2, 1000)
        self.assertEqual(inter.rgba, (255, 255, 255, 255))
    
    
    def test_create(self):
        """Tests whether create method works correctly."""
        
        # test from color
        color = pero.Color.create(pero.Color(255, 0, 0, 255))
        self.assertEqual(color.rgba, (255, 0, 0, 255))
        
        # test from name
        color = pero.Color.create('Red')
        self.assertEqual(color.rgba, (255, 0, 0, 255))
        
        # test from tuple
        color = pero.Color.create((170, 187, 204, 221))
        self.assertEqual(color.rgba, (170, 187, 204, 221))
        
        # test from hex
        color = pero.Color.create("#aabbccdd")
        self.assertEqual(color.rgba, (170, 187, 204, 221))
    
    
    def test_from_name(self):
        """Tests whether named colors can be accessed."""
        
        # test from name
        color = pero.Color.from_name('Red')
        color = pero.Color.from_name('red')
        
        # test from class
        color = pero.Color.Red
        color = pero.Color.red
        
        # test from lib
        color = pero.COLORS.Red
        color = pero.COLORS.red
        
        # test from module
        color = pero.colors.Red
    
    
    def test_from_hex(self):
        """Tests whether color is correctly created from hex."""
        
        # test all channels
        color = pero.Color("#aabbccdd")
        self.assertEqual(color.rgba, (170, 187, 204, 221))
        
        color = pero.Color("#abcd")
        self.assertEqual(color.rgba, (170, 187, 204, 221))
        
        # test without alpha
        color = pero.Color("#aabbcc")
        self.assertEqual(color.rgba, (170, 187, 204, 255))
        
        color = pero.Color("#abc")
        self.assertEqual(color.rgba, (170, 187, 204, 255))
    
    
    def test_name_registering(self):
        """Tests whether new named colors are registered."""
        
        # init named color
        model = pero.Color(0, 0, 0, 0, name="MyTransparent")
        
        # test from name
        color = pero.Color.from_name('MyTransparent')
        self.assertTrue(color is model)
        
        color = pero.Color.from_name('mytransparent')
        self.assertTrue(color is model)
        
        # test from class
        color = pero.Color.MyTransparent
        self.assertTrue(color is model)
        
        color = pero.Color.mytransparent
        self.assertTrue(color is model)
        
        # test from lib
        color = pero.COLORS.MyTransparent
        self.assertTrue(color is model)
        
        color = pero.COLORS.mytransparent
        self.assertTrue(color is model)


# run test case
if __name__ == "__main__":
    unittest.main(verbosity=2)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import unittest
import pero


class TestCase(unittest.TestCase):
    """Test case for engineering formatter."""
    
    
    def test_standalone(self):
        """Tests whether formatter works as standalone tool."""
        
        formatter = pero.EngFormatter()
        
        self.assertEqual(formatter.format(234.5678e6), "235 M")
        self.assertEqual(formatter.format(234.5678e3), "235 k")
        self.assertEqual(formatter.format(234.5678), "235")
        self.assertEqual(formatter.format(234.5678e-3), "235 m")
        self.assertEqual(formatter.format(234.5678e-6), "235 u")
    
    
    def test_units(self):
        """Tests whether units are added correctly."""
        
        formatter = pero.EngFormatter(units="Hz")
        
        self.assertEqual(formatter.format(234.5678e6), "235 MHz")
        self.assertEqual(formatter.format(234.5678e3), "235 kHz")
        self.assertEqual(formatter.format(234.5678), "235 Hz")
        self.assertEqual(formatter.format(234.5678e-3), "235 mHz")
        self.assertEqual(formatter.format(234.5678e-6), "235 uHz")
    
    
    def test_places(self):
        """Tests whether places works correctly."""
        
        formatter = pero.EngFormatter(places=0)
        
        self.assertEqual(formatter.format(234.5678e6), "235 M")
        self.assertEqual(formatter.format(234.5678e3), "235 k")
        self.assertEqual(formatter.format(234.5678), "235")
        self.assertEqual(formatter.format(234.5678e-3), "235 m")
        self.assertEqual(formatter.format(234.5678e-6), "235 u")
        
        formatter = pero.EngFormatter(places=2)
        
        self.assertEqual(formatter.format(234.5678e6), "234.57 M")
        self.assertEqual(formatter.format(234.5678e3), "234.57 k")
        self.assertEqual(formatter.format(234.5678), "234.57")
        self.assertEqual(formatter.format(234.5678e-3), "234.57 m")
        self.assertEqual(formatter.format(234.5678e-6), "234.57 u")
    
    
    def test_precision(self):
        """Tests whether precision works correctly."""
        
        formatter = pero.EngFormatter(precision=100)
        
        self.assertEqual(formatter.format(234.5678e6), "234.5678 M")
        self.assertEqual(formatter.format(234.5678e3), "234.6 k")
        self.assertEqual(formatter.format(234.5678), "235")
        self.assertEqual(formatter.format(234.5678e-3), "235 m")
        self.assertEqual(formatter.format(234.5678e-6), "235 u")
        
        formatter = pero.EngFormatter(precision=0.1)
        
        self.assertEqual(formatter.format(234.5678e6), "234.5678000 M")
        self.assertEqual(formatter.format(234.5678e3), "234.5678 k")
        self.assertEqual(formatter.format(234.5678), "234.6")
        self.assertEqual(formatter.format(234.5678e-3), "235 m")
        self.assertEqual(formatter.format(234.5678e-6), "235 u")
    
    
    def test_domain(self):
        """Tests whether domain works correctly."""
        
        formatter = pero.EngFormatter(domain=1e4)
        
        self.assertEqual(formatter.format(234.5678e6), "234568 k")
        self.assertEqual(formatter.format(234.5678e3), "235 k")
        self.assertEqual(formatter.format(234.5678), "0 k")
        self.assertEqual(formatter.format(234.5678e-3), "0 k")
        self.assertEqual(formatter.format(234.5678e-6), "0 k")
        
        formatter = pero.EngFormatter(domain=1e2)
        
        self.assertEqual(formatter.format(234.5678e6), "234567800")
        self.assertEqual(formatter.format(234.5678e3), "234568")
        self.assertEqual(formatter.format(234.5678), "235")
        self.assertEqual(formatter.format(234.5678e-3), "0")
        self.assertEqual(formatter.format(234.5678e-6), "0")
    
    
    def test_digits(self):
        """Tests whether digits works correctly."""
        
        # check domain and precision
        formatter = pero.EngFormatter(domain=1e4, precision=100)
        
        self.assertEqual(formatter.format(234.5678e6), "234567.8 k")
        self.assertEqual(formatter.format(234.5678e3), "234.6 k")
        self.assertEqual(formatter.format(234.5678), "0.2 k")
        self.assertEqual(formatter.format(234.5678e-3), "0.0 k")
        self.assertEqual(formatter.format(234.5678e-6), "0.0 k")
        
        # check domain, precision and places
        formatter = pero.EngFormatter(domain=1e4, precision=100, places=2)
        
        self.assertEqual(formatter.format(234.5678e6), "234567.80 k")
        self.assertEqual(formatter.format(234.5678e3), "234.57 k")
        self.assertEqual(formatter.format(234.5678), "0.23 k")
        self.assertEqual(formatter.format(234.5678e-3), "0.00 k")
        self.assertEqual(formatter.format(234.5678e-6), "0.00 k")
    
    
    def test_suffix(self):
        """Tests whether suffix works correctly."""
        
        # standalone formatter
        formatter = pero.EngFormatter(hide_suffix=True)
        self.assertEqual(formatter.suffix(), "")
        
        self.assertEqual(formatter.format(234.5678e6), "235")
        self.assertEqual(formatter.suffix(), " (M)")
        
        self.assertEqual(formatter.format(234.5678e3), "235")
        self.assertEqual(formatter.suffix(), " (k)")
        
        self.assertEqual(formatter.format(234.5678), "235")
        self.assertEqual(formatter.suffix(), "")
        
        # formatter with domain
        formatter = pero.EngFormatter(hide_suffix=True, domain=1e4)
        self.assertEqual(formatter.suffix(), " (k)")
        
        self.assertEqual(formatter.format(234.5678e6), "234568")
        self.assertEqual(formatter.suffix(), " (k)")
        
        self.assertEqual(formatter.format(234.5678e3), "235")
        self.assertEqual(formatter.suffix(), " (k)")
        
        self.assertEqual(formatter.format(234.5678), "0")
        self.assertEqual(formatter.suffix(), " (k)")
    
    
    def test_convert(self):
        """Tests whether convert works correctly."""
        
        formatter = pero.EngFormatter(domain=1e4)
        
        self.assertEqual(formatter.convert(234.5678e6), 234567.8)
        self.assertEqual(formatter.convert(234.5678e4), 2345.678)
        self.assertEqual(formatter.convert(234.5678), 0.2345678)
        self.assertEqual(formatter.convert(234.5678e-4), 2.345678e-5)
        self.assertEqual(formatter.convert(234.5678e-6), 2.345678e-7)
    
    
    def test_invert(self):
        """Tests whether invert works correctly."""
        
        formatter = pero.EngFormatter(domain=1e4)
        
        self.assertEqual(formatter.invert(234567.8), 234.5678e6)
        self.assertEqual(formatter.invert(2345.678), 234.5678e4)
        self.assertEqual(formatter.invert(0.2345678), 234.5678)
        self.assertEqual(formatter.invert(2.345678e-5), 234.5678e-4)
        self.assertEqual(formatter.invert(2.345678e-7), 234.5678e-6)


# run test case
if __name__ == "__main__":
    unittest.main(verbosity=2)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.


class Library(object):
    """
    Represents a named library of available items used to provide library of
    predefined colors, palettes and gradients. Each added item is registered
    using its original name as well as the lowercase version. 
    """
    
    
    def __init__(self):
        """Initializes a new instance of Library."""
        
        self._items = {}
        self._names = {}


    def __len__(self):
        """Gets number of available colors."""
    
        return len(self._items)


    def __contains__(self, key):
        """Checks whether item exists."""
        
        return key in self._names
    
    
    def __getattr__(self, key):
        """Gets registered item by its name."""
        
        if key in self._names:
            return self._items[self._names[key]]
        
        raise AttributeError(key)
    
    
    def __getitem__(self, key):
        """Gets registered item by its name."""
        
        if key in self._names:
            return self._items[self._names[key]]
        
        raise KeyError(key)
    
    
    def __iter__(self):
        """Gets items iterator."""
        
        return self._items.values().__iter__()
    
    
    def add(self, item):
        """Adds new item."""
        
        self._items[item.name] = item
        self._names[item.name] = item.name
        self._names[item.name.lower()] = item.name



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import unittest
import pero


class TestCase(unittest.TestCase):
    """Test case for fixed values ticker."""
    
    
    def test_major_ticks(self):
        """Tests whether major ticks are calculated correctly."""
        
        ticker = pero.FixTicker(major_values=(2, 4, 8, 16))
        
        # test within range
        ticker(start=5, end=10)
        ticks = ticker.major_ticks()
        self.assertEqual(ticks, (8,))
        
        # test outside range
        ticker(start=100, end=1000)
        ticks = ticker.major_ticks()
        self.assertEqual(ticks, ())
        
        # test cross range
        ticker(start=5, end=1000)
        ticks = ticker.major_ticks()
        self.assertEqual(ticks, (8, 16))
        
        # test flipped
        ticker(start=1000, end=5)
        ticks = ticker.major_ticks()
        self.assertEqual(ticks, (8, 16))
    
    
    def test_minor_ticks(self):
        """Tests whether minor ticks are calculated correctly."""
        
        ticker = pero.FixTicker(minor_values=(2, 3, 4, 6, 8, 12, 16))
        
        # test within range
        ticker(start=5, end=10)
        ticks = ticker.minor_ticks()
        self.assertEqual(ticks, (6, 8))
        
        # test outside range
        ticker(start=100, end=1000)
        ticks = ticker.minor_ticks()
        self.assertEqual(ticks, ())
        
        # test cross range
        ticker(start=5, end=1000)
        ticks = ticker.minor_ticks()
        self.assertEqual(ticks, (6, 8, 12, 16))
        
        # test flipped
        ticker(start=1000, end=5)
        ticks = ticker.minor_ticks()
        self.assertEqual(ticks, (6, 8, 12, 16))


# run test case
if __name__ == "__main__":
    unittest.main(verbosity=2)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

from .. enums import *
from .. properties import *
from .. colors import Gradient, Transparent
from . interpols import *
from . scale import Scale


class GradientScale(Scale):
    """
    This type of scale tool is used to convert continuous numerical input values
    into color gradient. This is typically used to convert real data units
    defined by 'in_range' property into a color gradient e.g. to plot data
    series. This type of scale does not provide the 'invert' method.
    
    The conversion is done in two steps: normalizing given value within the
    input range and converting the normalized value within the output gradient.
    To provide a specific way of input range normalization, appropriate
    'normalizer' must be specified. There are already some predefined gradient
    scales for convenient use like GradientLinScale, GradientLogScale and
    GradientPowScale.
    
    The output colors should be provided either as a list of colors or palette
    into the 'out_range', from which the gradient is created automatically.
    
    Properties:
        
        in_range: (float, float)
            Specifies the start and end value of the input range.
        
        out_range: pero.Gradient, pero.Palette, (color,), str
            Specifies the output gradient as a sequence of supported pero.Color
            definitions, pero.Palette, pero.Gradient or registered palette or
            gradient name.
        
        normalizer: pero.Interpol
            Specifies the normalization interpolator.
    """
    
    in_range = TupleProperty((), intypes=(int, float), dynamic=False)
    out_range = GradientProperty(UNDEF, dynamic=False)
    
    normalizer = Property(UNDEF, types=(Interpol,), dynamic=False)
    
    
    def __init__(self, **overrides):
        """Initializes a new instance of GradientScale."""
        
        # init interpolator
        if 'normalizer' not in overrides:
            overrides['normalizer'] = LinInterpol()
        
        # init base
        super().__init__(**overrides)
        
        # init gradient
        self._gradient = None
        self._update_gradient()
        
        # bind events
        self.bind(EVT_PROPERTY_CHANGED, self._on_gradient_scale_property_changed)
    
    
    def scale(self, value, *args, **kwargs):
        """
        Returns corresponding output color for given input value.
        
        Args:
            value: float or (float,)
                Input value to be scaled.
        
        Returns:
            pero.Color
                Output color.
        """
        
        # apply array scaling
        if isinstance(value, (numpy.ndarray, list, tuple)):
            return tuple(map(self.scale, value))
        
        # normalize value
        norm = self.normalizer.normalize(value, self.in_range[0], self.in_range[1])
        
        # convert normalized value into color
        return self._gradient.color_at(norm)
    
    
    def _update_gradient(self):
        """Updates gradient by current out_range."""
        
        if self.out_range is UNDEF:
            self._gradient = Gradient((Transparent, Transparent))
        
        else:
            self._gradient = self.out_range.normalized(0, 1)
    
    
    def _on_gradient_scale_property_changed(self, evt=None):
        """Called after a property has changed."""
        
        # update gradient
        if evt is None or evt.name == 'out_range':
            self._update_gradient()


class GradientLinScale(GradientScale):
    """Gradient scale with linear transformation."""
    
    
    def __init__(self, **overrides):
        """Initializes a new instance of GradientLinScale."""
        
        # init normalizer
        overrides['normalizer'] = LinInterpol()
        
        # init base
        super().__init__(**overrides)
        
        # lock normalizer
        self.lock_property('normalizer')


class GradientLogScale(GradientScale):
    """Gradient scale with logarithmic transformation."""
    
    
    def __init__(self, **overrides):
        """Initializes a new instance of GradientLogScale."""
        
        # init normalizer
        overrides['normalizer'] = LogInterpol()
        
        # init base
        super().__init__(**overrides)
        
        # lock normalizer
        self.lock_property('normalizer')


class GradientPowScale(GradientScale):
    """
    Gradient scale with exponential transformation.
    
    Properties:
        
        power: int or float
            Specifies the exponent to be used.
    """
    
    power = FloatProperty(1, dynamic=False)
    
    
    def __init__(self, **overrides):
        """Initializes a new instance of GradientPowScale."""
        
        # init normalizer
        overrides['normalizer'] = PowInterpol()
        
        # init base
        super().__init__(**overrides)
        
        # lock normalizer
        self.lock_property('normalizer')
        
        # set power
        self.normalizer.power = self.power
        
        # bind events
        self.bind(EVT_PROPERTY_CHANGED, self._on_pow_scale_property_changed)
    
    
    def _on_pow_scale_property_changed(self, evt=None):
        """Called after a property has changed."""
        
        # check power
        if evt is None or evt.name == 'power':
            self.normalizer.power = self.power



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

# import main objects
from . enums import *
from . canvas import UICanvas
from . view import UIView
from . viewer import UIViewer
from . export import export, show



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.


class Event(object):
    """
    Abstract base class for various types of events. Each derived event class
    has to specify its unique TYPE property at least.
    """
    
    TYPE = None
    
    
    def __init__(self, **kwargs):
        """Initializes a new instance of Event."""
        
        self._canceled = False
        
        # set given arguments
        for name, value in kwargs.items():
            if hasattr(self, name):
                setattr(self, name, value)
            else:
                raise AttributeError("Attribute not found! --> %s" % name)
    
    
    def __str__(self):
        """Gets standard string representation."""
        
        # add type
        text = "%s" % self.TYPE
        
        # mark as cancelled
        if self._canceled:
            text += "(canceled)"
        
        return text
    
    
    def __repr__(self):
        """Gets debug string representation."""
        
        return "%s(%s)" % (self.__class__.__name__, self.__str__())
    
    
    def cancel(self):
        """
        Sets current event as canceled to prevent following subscribers to be
        called.
        """
        
        self._canceled = True
    
    
    def resume(self):
        """
        Sets current event as not canceled to allow following subscribers to be
        called.
        """
        
        self._canceled = False
    
    
    def is_canceled(self):
        """
        Gets the value indicating if the event has been canceled.
        
        Returns:
            bool
                Canceled flag.
        """
        
        return self._canceled



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

from .. enums import *
from .. properties import *
from . glyph import Glyph


class Framer(Glyph):
    """
    Defines a utility glyph visualizing the details of given rectangle. If the
    'width' or 'height' property is set to pero.UNDEF, corresponding value is
    taken directly from given canvas when drawn.
    
    Properties:
        
        show_label: bool, callable
            Specifies whether label should be shown.
        
        show_size: bool, callable
            Specifies whether width and height should be shown.
        
        label: str, callable
            Specifies the label text to show.
        
        x: int, float or callable
            Specifies the x-coordinate of the top-left corner.
        
        y: int, float or callable
            Specifies the y-coordinate of the top-left corner.
        
        width: int, float or callable
            Specifies the width. If not defined, the value is taken
            automatically from given canvas.
        
        height: int, float or callable
            Specifies the height. If not defined, the value is taken
            automatically from given canvas.
        
        line properties:
            Includes pero.LineProperties to specify the path outline.
        
        fill properties:
            Includes pero.FillProperties to specify the path fill.
        
        text properties:
            Includes pero.TextProperties to specify the label properties.
            Some of them (e.g. alignment, angle, baseline) are set automatically
            and cannot be changed.
    """
    
    show_label = BoolProperty(True)
    show_size = BoolProperty(True)
    
    label = StringProperty(UNDEF)
    
    x = NumProperty(0)
    y = NumProperty(0)
    width = NumProperty(UNDEF)
    height = NumProperty(UNDEF)
    
    outline = Include(LineProperties, line_color=None)
    fill = Include(FillProperties, fill_color="#ccc")
    font = Include(TextProperties)
    
    
    def draw(self, canvas, source=UNDEF, **overrides):
        """Uses given canvas to draw the frame."""
        
        # check if visible
        if not self.is_visible(source, overrides):
            return
        
        # get properties
        tag = self.get_property('tag', source, overrides)
        show_label = self.get_property('show_label', source, overrides)
        show_size = self.get_property('show_size', source, overrides)
        label = self.get_property('label', source, overrides)
        x = self.get_property('x', source, overrides)
        y = self.get_property('y', source, overrides)
        width = self.get_property('width', source, overrides)
        height = self.get_property('height', source, overrides)
        
        # get size from canvas
        if width is UNDEF:
            width = canvas.viewport.width - x
        if height is UNDEF:
            height = canvas.viewport.height - y
        
        # get full label
        labels = []
        
        if show_label and label:
            labels.append("%s" % label)
        if show_size:
            labels.append("%d x %d" % (width, height))
        
        label = "\n".join(labels)
        
        # set text
        canvas.set_text_by(self, source=source, overrides=overrides)
        canvas.text_align = TEXT_ALIGN_CENTER
        canvas.text_base = TEXT_BASE_MIDDLE
        
        # set pen and brush
        canvas.set_pen_by(self, source=source, overrides=overrides)
        canvas.set_brush_by(self, source=source, overrides=overrides)
        
        # start drawing group
        canvas.group(tag, "framer")
        
        # draw
        canvas.draw_rect(x, y, width, height)
        canvas.draw_line(x, y, x+width, y+height)
        canvas.draw_line(x, y+height, x+width, y)
        canvas.draw_text(label, x+0.5*width, y+0.5*height)
        
        # end drawing group
        canvas.ungroup()



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

from .. enums import *
from .. properties import *
from .. formatters import TimeFormatter
from .. formatters.utils import *
from . utils import *
from . ticker import Ticker


class TimeTicker(Ticker):
    """
    This type of ticker generates nice looking ticks and labels for time data
    assuming all given data are in seconds. The major and the minor steps are
    calculated automatically according to current range, however, both can be
    specified if needed. The label 'formatter' is by default set to
    pero.TimeFormatter, but can be changed if needed.
    
    Properties:
        
        major_step: int, float or UNDEF
            Specifies the fixed step in seconds to be used for major ticks and
            labels. If not provided the value is determined automatically base
            on current range.
        
        major_count: int
            Specifies the expected number of major ticks to be generated for
            current range. The final number of ticks should be close but not
            necessary the same.
        
        minor_step: int, float or UNDEF
            Specifies the fixed step in seconds to be used for minor ticks. If
            not provided the value is determined automatically base on major
            ticks range.
        
        minor_count: int
            Specifies the expected number of minor ticks to be generated between
            two major ticks. The final number of ticks should be close but not
            necessary the same.
    """
    
    major_step = NumProperty(UNDEF, dynamic=False)
    major_count = IntProperty(7, dynamic=False)
    
    minor_step = NumProperty(UNDEF, dynamic=False)
    minor_count = IntProperty(5, dynamic=False)
    
    
    def __init__(self, **overrides):
        """Initializes a new instance of TimeTicker."""
        
        # init formatter
        if 'formatter' not in overrides:
            overrides['formatter'] = TimeFormatter(rounding=ROUNDING.HALFUP)
        
        # init base
        super().__init__(**overrides)
    
    
    def make_ticks(self):
        """
        Generates ticks according to current settings.
        
        Returns:
            (float,), (float,)
                Generated major and minor ticks.
        """
        
        # get range
        start, end = self.start, self.end
        domain = abs(end - start)
        
        # calc major step size
        major_step = self.major_step
        if major_step is UNDEF:
            major_step = self._calc_step(domain, self.major_count)
        
        # calc minor step size
        minor_step = self.minor_step
        if minor_step is UNDEF:
            minor_step = self._calc_step(major_step, self.minor_count)
        
        # make ticks
        major_ticks = make_lin_ticks(start, end, major_step)
        minor_ticks = make_lin_ticks(start, end, minor_step)
        
        # update formatter
        self.formatter.precision = abs(major_step)
        
        return major_ticks, minor_ticks
    
    
    def beautify(self, start, end):
        """
        Calculates visually nice range to cover given range.
        
        Args:
            start: int or float
                Minimum value of the range.
            
            end: int or float
                Maximum value of the range.
        
        Returns:
            (float, float)
                New extended range.
        """
        
        # check order
        flip = False
        if start > end:
            start, end = end, start
            flip = True
        
        # calc major step size
        step = self.major_step
        if step is UNDEF:
            step = self._calc_step(abs(end - start), self.major_count)
        
        # extend range
        start = math.floor(start / step) * step
        end = math.ceil(end / step) * step
        
        return (start, end) if not flip else (end, start)
    
    
    def _calc_step(self, domain, count):
        """Calculates minor step size."""
        
        # split time
        parts = split_value(domain, SPLITS_TIME)
        
        # get split
        if parts["h"] or parts["m"] or parts["s"]:
            splits = (30, 15, 10, 5, 2, 1)
            base = 60
        
        else:
            splits = (5, 2, 1)
            base = 10
        
        # calc step size
        step = calc_step_size(domain, count, splits, base)
        
        return step



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.


def split_value(value, splits):
    """
    Splits given time in seconds into available units.
    
    Args:
        value: float or int
            Value to split.
        
        splits: {str: float}
            Split names and thresholds values.
    
    Returns:
        {str: float}
            Value for each available split.
    """
    
    # get available units
    units = sorted(splits.items(), key=lambda d: d[1], reverse=True)
    
    # init parts
    parts = {u[0]: 0.0 for u in units}
    
    # split time
    rest = value
    for unit, f in units:
        count, rest = divmod(rest, f)
        parts[unit] = count
    
    return parts



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import py5
import numpy
from ... properties import *
from ... colors import Transparent, Black
from ... drawing import Canvas, Path, Matrix, ClipState
from . enums import *


class Py5Canvas(Canvas):
    """Wrapper for Py5 drawing graphics."""
    
    
    def __init__(self, pg, context=py5, **overrides):
        """
        Initializes a new instance of Py5Canvas.
        
        Args:
            pg: py5.Py5Graphics
                Sketch object to draw into.
            
            context: py5 module or py5.Sketch
                Context in which the drawing is done.
            
            overrides: str:any pairs
                Value overwrites for specific properties.
        """
        
        # init buffers
        self._pg = pg
        self._context = context
        self._clipping = []
        
        self._line_color = None
        self._fill_color = None
        self._for_color = Black
        self._bgr_color = None
        
        # init size
        if 'width' not in overrides:
            overrides['width'] = pg.width
        
        if 'height' not in overrides:
            overrides['height'] = pg.height
        
        # init base
        super().__init__(**overrides)
        
        # bind events
        self.bind(EVT_PEN_CHANGED, self._update_pen)
        self.bind(EVT_BRUSH_CHANGED, self._update_brush)
        self.bind(EVT_TEXT_CHANGED, self._update_text)
    
    
    def get_line_size(self, text):
        """
        Gets width and height of a single text line using current text settings.
        
        Args:
            text: str
                Text for which the size should be calculated.
        
        Returns:
            (float, float)
                Line width and height.
        """
        
        # check text
        if not text:
            return 0, 0
        
        # get size
        width = self._pg.text_width(text)
        height = self._pg.text_ascent() + self._pg.text_descent()
        
        return width, height
    
    
    def draw_circle(self, x, y, radius):
        """
        Draws a circle of specified radius centered around given coordinates.
        
        Args:
            x: int or float
                X-coordinate of the center.
            
            y: int or float
                Y-coordinate of the center.
            
            radius: int or float
                Radius of the circle.
        """
        
        # apply scaling and offset
        x = self._scale * (x + self._offset[0])
        y = self._scale * (y + self._offset[1])
        radius = self._scale * radius
        
        # draw circle
        self._pg.ellipse_mode(py5.CENTER)
        self._pg.circle(x, y, 2*radius)
    
    
    def draw_ellipse(self, x, y, width, height):
        """
        Adds an ellipse centered at given position and fitting into the size.
        
        Args:
            x: int or float
                X-coordinate of the center.
            
            y: int or float
                Y-coordinate of the center.
            
            width: int or float
                Full width of the ellipse.
            
            height: int or float
                Full height of the ellipse.
        """
        
        # apply scaling and offset
        x = self._scale * (x + self._offset[0])
        y = self._scale * (y + self._offset[1])
        width = self._scale * width
        height = self._scale * height
        
        # check size
        if not width or not height:
            return
        
        # draw ellipse
        self._pg.ellipse_mode(py5.CENTER)
        self._pg.ellipse(x, y, width, height)
    
    
    def draw_line(self, x1, y1, x2, y2):
        """
        Draws a line between two points.
        
        Args:
            x1: int or float
                X-coordinate of the line start.
            
            y1: int or float
                Y-coordinate of the line start.
            
            x2: int or float
                X-coordinate of the line end.
            
            y2: int or float
                Y-coordinate of the line end.
        """
        
        # apply scaling and offset
        x1 = self._scale * (x1 + self._offset[0])
        y1 = self._scale * (y1 + self._offset[1])
        x2 = self._scale * (x2 + self._offset[0])
        y2 = self._scale * (y2 + self._offset[1])
        
        # draw line
        self._pg.line(x1, y1, x2, y2)
    
    
    def draw_lines(self, points):
        """
        Draws continuous open line using sequence of points.
        
        Args:
            points: ((float, float),)
                Sequence of x,y coordinates of the points.
        """
        
        # check points
        if len(points) < 2:
            return
        
        # apply scaling and offset
        points = (numpy.array(points) + self._offset) * numpy.array((self._scale, self._scale))
        
        # create coords
        points = numpy.concatenate((points[:-1], points[1:]), axis=1)
        
        # draw lines
        self._pg.lines(points)
    
    
    def draw_path(self, path):
        """
        Draws the path using current pen and brush.
        
        Args:
            path: pero.Path
                Path to be drawn.
        """
        
        # apply scaling and offset
        matrix = Matrix()
        matrix.translate(self._offset[0], self._offset[1])
        matrix.scale(self._scale, self._scale)
        path = path.transformed(matrix)
        
        # draw individual sub-paths
        for sub in path.split():
            
            # begin shape
            self._pg.begin_shape()
            end = py5.OPEN
            
            # draw path
            for command in sub.commands():
                
                # get data
                key = command[0]
                values = command[1:]
                
                # close
                if key == PATH_CLOSE:
                    end = py5.CLOSE
                
                # move to
                elif key == PATH_MOVE:
                    self._pg.vertex(*values[0:2])
                
                # line to
                elif key == PATH_LINE:
                    self._pg.vertex(*values[0:2])
                
                # curve to
                elif key == PATH_CURVE:
                    self._pg.bezier_vertex(*values)
            
            # end shape
            self._pg.end_shape(end)
    
    
    def draw_rect(self, x, y, width, height, radius=None):
        """
        Draws a rectangle specified by given top left corner and size and
        optional round corners specified as a single value or individual value
        for each corner starting from top-left.
        
        Args:
            x: int or float
                X-coordinate of the top left corner.
            
            y: int or float
                Y-coordinate of the top left corner.
            
            width: int or float
                Full width of the rectangle.
            
            height: int or float
                Full height of the rectangle.
            
            radius: int, float, (int,int,int,int) or (float,float,float,float)
                Radius of curved corners.
        """
        
        # get radius
        if isinstance(radius, (int, float)) and radius != 0:
            radius = (radius, radius, radius, radius)
        
        # draw as path if having round corners
        if radius:
            path = Path()
            path.rect(x, y, width, height, radius)
            self.draw_path(path)
            return
        
        # apply scaling and offset
        x = self._scale * (x + self._offset[0])
        y = self._scale * (y + self._offset[1])
        width = self._scale * width
        height = self._scale * height
        
        # draw rect
        self._pg.rect_mode(py5.CORNER)
        self._pg.rect(x, y, width, height)
    
    
    def draw_text(self, text, x, y, angle=0):
        """
        Draws a text string anchored at specified point using current text
        settings.
        
        Args:
            text: str
                Text to be drawn.
            
            x: int or float
                X-coordinate of the text anchor.
            
            y: int or float
                Y-coordinate of the text anchor.
            
            angle: int or float
                Text angle in radians.
        """
        
        # set colors
        self._pg.no_stroke()
        if self._for_color:
            self._pg.fill(self._for_color.hex)
        
        # get full size
        full_width, full_height = self.get_text_size(text)
        
        # apply angle transformation
        if angle:
            
            shift_x = self._scale * (x + self._offset[0])
            shift_y = self._scale * (y + self._offset[1])
            
            self._pg.translate(shift_x, shift_y)
            self._pg.rotate(angle)
            
            x = 0
            y = 0
        
        # split lines
        lines = [text]
        if self.text_split and self.text_splitter:
            lines = text.split(self.text_splitter)
        
        # draw lines
        for i, line in enumerate(lines):
            
            # get line size
            line_width, line_height = self.get_line_size(line)
            line_width /= self._scale
            line_height /= self._scale
            
            # init offset
            x_offset = 0
            y_offset = 0
            
            # adjust alignment
            if self.text_align == TEXT_ALIGN_CENTER:
                x_offset -= 0.5 * line_width
            
            elif self.text_align == TEXT_ALIGN_RIGHT:
                x_offset -= line_width
            
            # adjust baseline
            if self.text_base == TEXT_BASE_MIDDLE:
                y_offset -= 0.5 * full_height
            
            elif self.text_base == TEXT_BASE_BOTTOM:
                y_offset -= full_height
            
            # add line offset
            y_offset += i * line_height * (1 + self.text_spacing)
            
            # apply scaling and offset
            if angle:
                line_x = self._scale * x_offset
                line_y = self._scale * y_offset
            else:
                line_x = self._scale * (x + x_offset + self._offset[0])
                line_y = self._scale * (y + y_offset + self._offset[1])
            
            # draw background
            if self._bgr_color:
                
                bgr_x = line_x
                bgr_y = line_y
                bgr_width = line_width * self._scale
                bgr_height = line_height * self._scale
                
                self._pg.fill(self._bgr_color.hex)
                self._pg.rect(bgr_x, bgr_y, bgr_width, bgr_height)
                
                if self._for_color:
                    self._pg.fill(self._for_color.hex)
            
            # draw text
            self._pg.text_align(py5.LEFT, py5.TOP)
            self._pg.text(line, line_x, line_y)
        
        # revert colors
        if self._line_color:
            self._pg.stroke(self._line_color.hex)
        
        if self._fill_color:
            self._pg.fill(self._fill_color.hex)
        
        # revert angle transformation
        if angle:
            self._pg.rotate(-angle)
            self._pg.translate(-shift_x, -shift_y)
    
    
    def clip(self, path):
        """
        Sets clipping path as intersection with current one.
        
        Current implementation does not truly support paths so it uses just the
        rectangular bounding box of the given path.
        
        Args:
            path: pero.Path
                Path to be used for clipping.
        
        Returns:
            pero.ClipState
                Clipping state context manager.
        """
        
        # apply scaling and offset
        matrix = Matrix()
        matrix.translate(self._offset[0], self._offset[1])
        matrix.scale(self._scale, self._scale)
        path = path.transformed(matrix)
        
        # set clipping
        rect = path.bbox().rect
        self._pg.image_mode(py5.CORNER)
        self._pg.clip(*rect)
        
        # remember clipping
        self._clipping.append(rect)
        
        # return state
        return ClipState(self)
    
    
    def unclip(self):
        """Removes last clipping path while keeping previous if any."""
        
        # remove clip
        self._pg.no_clip()
        
        # remove from stack
        if self._clipping:
            del self._clipping[-1]
        
        # re-apply previous
        if self._clipping:
            self._pg.image_mode(py5.CORNER)
            self._pg.clip(*self._clipping[-1])
    
    
    def _update_pen(self, evt=None):
        """Updates pen with current properties."""
        
        # get property name
        prop_name = evt.name if evt is not None else None
        
        # update color
        if prop_name is None or prop_name in ('line_color', 'line_alpha'):
            color = ColorProperties.get_color(self, "line_")
            if color is not UNDEF:
                self._line_color = color
                self._pg.stroke(color.hex)
        
        # update width
        if prop_name is None or prop_name in ('line_width', 'line_scale'):
            line_width = self.line_width
            if line_width is not UNDEF:
                self._pg.stroke_weight(line_width)
        
        # update cap
        if prop_name is None or prop_name == 'line_cap':
            line_cap = self.line_cap
            if line_cap is not UNDEF:
                self._pg.stroke_cap(PY5_LINE_CAP[line_cap])
        
        # update join
        if prop_name is None or prop_name == 'line_join':
            line_join = self.line_join
            if line_join is not UNDEF:
                self._pg.stroke_join(PY5_LINE_JOIN[line_join])
    
    
    def _update_brush(self, evt=None):
        """Updates brush with current properties."""
        
        # get color
        color = ColorProperties.get_color(self, "fill_")
        
        # set transparent fill
        if self.fill_style == FILL_STYLE_TRANS:
            self._fill_color = Transparent
            self._pg.fill(self._fill_color.hex)
        
        # set color fill
        elif color is not UNDEF:
            self._fill_color = color
            self._pg.fill(self._fill_color.hex)
    
    
    def _update_text(self, evt=None):
        """Updates text with current properties."""
        
        # get property name
        prop_name = evt.name if evt is not None else None
        
        # update font
        if prop_name is None or prop_name in ('font_name', 'font_family', 'font_style', 'font_weight'):
            font_size = self.font_size or 10
            font = self.get_font()
            font = self._context.create_font(font.path, font_size * self.font_scale)
            self._pg.text_font(font)
        
        # update font size
        if prop_name is None or prop_name in ('font_size', 'font_scale'):
            font_size = self.font_size
            if font_size is not UNDEF:
                self._pg.text_size(font_size * self.font_scale)
        
        # update foreground color
        if prop_name is None or prop_name in ('text_color', 'text_alpha'):
            color = ColorProperties.get_color(self, "text_")
            if color is not UNDEF:
                self._for_color = color
        
        # update background color
        if prop_name is None or prop_name in ('text_bgr_color', 'text_bgr_alpha'):
            color = ColorProperties.get_color(self, "text_bgr_")
            if color is not UNDEF:
                self._bgr_color = color



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import ui
from objc_util import ObjCInstance
from time import time
from ... events import *
from .. view import View
from . enums import *
from . canvas import UICanvas

# define constants
_DTAP_DELAY = 0.3


class UIView(ui.View, View):
    """Wrapper for UI View."""
    
    
    def __init__(self):
        """Initializes a new instance of UIView."""
        
        # init base
        super(UIView, self).__init__()
        View.__init__(self)
        
        # init buffers
        self._dc_buffer = None
        self._dc_overlay = None
        
        # double-tap
        self._dtap_began = 0
        self._dtap_possible = False
    
    
    def draw_control(self):
        """Draws current control graphics."""
        
        # draw on buffer
        with ui.ImageContext(self.width, self.height) as ctx:
            
            # draw control
            if self.control is not None:
                canvas = UICanvas(width=self.width, height=self.height)
                self.control.draw(canvas)
            
            # get image
            self._dc_buffer = ctx.get_image()
        
        # reset overlay
        self._dc_overlay = None
        
        # update screen
        self.set_needs_display()
    
    
    def draw_overlay(self, func=None, view=None, **kwargs):
        """
        Draws cursor rubber band overlay.
        
        Specified function is expected to be called with a canvas as the first
        argument followed by given overrides (i.e. func(canvas, **kwargs)).
        If the 'func' parameter is set to None current overlay is cleared.
        
        Args:
            func: callable or None
                Drawing function to be called to draw the overlay. If set to
                None, current overlay will be cleared.
            
            view: (int, int, int, int) or None
                Rectangle defined as (x, y, width, height) used to shift origin
                of drawing canvas submitted to given drawing function.
                
            kwargs: str:any pairs
                Keyword arguments, which should be provided to the given drawing
                function.
        """
        
        # skip cleaning if empty already
        if func is None and self._dc_overlay is None:
            return
        
        # just clean
        if func is None:
            self._dc_overlay = None
            self.set_needs_display()
            return
        
        # draw on overlay
        with ui.ImageContext(self.width, self.height) as ctx:
            
            # init canvas
            canvas = UICanvas(width=self.width, height=self.height)
            
            # draw overlay
            if view is not None:
                with canvas.view(*view, relative=True):
                    func(canvas, **kwargs)
            else:
                func(canvas, **kwargs)
            
            # get image
            self._dc_overlay = ctx.get_image()
        
        # update screen
        self.set_needs_display()
    
    
    def layout(self):
        """Called when view is resized."""
        
        # make size event
        size_evt = SizeEvt(
            
            native = None,
            view = self,
            control = self.control,
            
            width = self.width,
            height = self.height)
        
        # fire event
        if self.control is not None:
            self.control.fire(size_evt)
    
    
    def draw(self):
        """Draws the view."""
        
        # draw buffer
        if self._dc_buffer is not None:
            self._dc_buffer.draw(0, 0, self.width, self.height)
        
        # draw overlay
        if self._dc_overlay is not None:
            self._dc_overlay.draw(0, 0, self.width, self.height)
    
    
    def touch_began(self, touch):
        """Called when a touch event begins."""
        
        # await double-tap
        if self._dtap_began:
            self._dtap_possible = True
        else:
            self._dtap_began = time()
            self._dtap_possible = False
        
        # create event
        self._on_touch(touch)
    
    
    def touch_ended(self, touch):
        """Called when a touch event ends."""
        
        # check double-tap
        if time() - self._dtap_began > _DTAP_DELAY:
            self._dtap_began = 0
            self._dtap_possible = False
        
        # create event
        self._on_touch(touch)
    
    
    def touch_moved(self, touch):
        """Called when a touch event moves."""
        
        # create event
        self._on_touch(touch)
    
    
    def _on_touch(self, touch):
        """Handles all touch events."""
        
        # get point
        point = Touch(
            id = touch.touch_id,
            x_pos = touch.location[0],
            y_pos = touch.location[1],
            x_prev = touch.prev_location[0],
            y_prev = touch.prev_location[1],
            force = ObjCInstance(touch).force(),
            state = UI_TOUCH_STATE[touch.phase])
        
        # init base event
        touch_evt = TouchEvt(
            
            native = touch,
            view = self,
            control = self.control,
            
            touches = [point],
            
            alt_down = False,
            cmd_down = False,
            ctrl_down = False,
            shift_down = False)
        
        # make specific event type
        if touch.phase == 'began':
            touch_evt = TouchStartEvt.from_evt(touch_evt)
        
        elif touch.phase == 'ended':
            if self._dtap_possible:
                self._dtap_began = 0
                self._dtap_possible = False
                touch_evt = TouchDTapEvt.from_evt(touch_evt)
            else:
                touch_evt = TouchEndEvt.from_evt(touch_evt)
        
        elif touch.phase == 'moved':
            touch_evt = TouchMoveEvt.from_evt(touch_evt)
        
        elif touch.phase == 'cancelled':
            touch_evt = TouchCancelEvt.from_evt(touch_evt)
        
        # fire event
        if self.control is not None:
            self.control.fire(touch_evt)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

from .. properties import *
from .. formatters import Formatter
from . ticker import Ticker


class FixTicker(Ticker):
    """
    This type of ticker provides user-specified fixed ticks. The label
    'formatter' is by default set to pero.Formatter, but can be changed if
    needed.
    
    Properties:
        
        major_values: tuple
            Specifies the values to be used as the major ticks and labels.
        
        minor_values: tuple
            Specifies the values to be used as the minor ticks.
    """
    
    major_values = TupleProperty((), dynamic=False)
    minor_values = TupleProperty((), dynamic=False)
    
    
    def __init__(self, **overrides):
        """Initializes a new instance of FixTicker."""
        
        # init formatter
        if 'formatter' not in overrides:
            overrides['formatter'] = Formatter()
        
        # init base
        super().__init__(**overrides)
    
    
    def make_ticks(self):
        """
        Generates ticks according to current settings.
        
        Returns:
            (float,), (float,)
                Generated major and minor ticks.
        """
        
        # get range
        start, end = self.start, self.end
        if start > end:
            start, end = end, start
        
        # get ticks in current range
        major_ticks = tuple(x for x in self.major_values if start <= x <= end)
        minor_ticks = tuple(x for x in self.minor_values if start <= x <= end)
        
        return major_ticks, minor_ticks



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import math
from .. enums import *
from .. colors import Color
from . typed import *
from . special import *
from . undefined import UNDEF
from . propset import PropertySet, Include, PROP_SPLITTER


class AngleProperties(PropertySet):
    """
    Collection of properties defining an angle value with its units.
    
    Properties:
        
        angle: int, float, callable, None or UNDEF
            Specifies the angle value.
        
        angle_units: pero.ANGLE, callable or UNDEF
            Specifies the angle units as any item from the pero.ANGLE enum.
    """
    
    angle = NumProperty(0, nullable=True)
    angle_units = EnumProperty(ANGLE_RAD, enum=ANGLE)
    
    
    @staticmethod
    def get_angle(prop_set, prefix="", units=ANGLE_RAD, source=UNDEF, overrides=None):
        """
        Gets current angle value directly converted to requested units.
        
        Args:
            prop_set: pero.PropertySet
                Property set from which to retrieve the angle.
            
            prefix: str
                Prefix applied to all angle properties.
            
            units: str
                Requested units of the angle as any item from the pero.ANGLE
                enum.
            
            source: any
                Data source to be used for retrieving the final value of
                callable properties.
            
            overrides: dict or None
                Highest priority properties to be used instead of current values.
        
        Returns:
            float or UNDEF
                Angle value in requested units or UNDEF if value or units are
                not defined.
        """
        
        # check type
        if not isinstance(prop_set, PropertySet):
            message = "Property set must be of type pero.PropertySet! -> %s" % type(prop_set)
            raise TypeError(message)
        
        # check prefix
        if prefix and prefix[-1] != PROP_SPLITTER:
            prefix = prefix + PROP_SPLITTER
        
        # get properties
        angle = prop_set.get_property(prefix+'angle', source, overrides)
        angle_units = prop_set.get_property(prefix+'angle_units', source, overrides)
        
        # check if None
        if angle is None or angle_units is None:
            return 0
        
        # check if defined
        if angle is UNDEF or angle_units is UNDEF:
            return UNDEF
        
        # check units
        if units == angle_units:
            return angle
        
        elif units == ANGLE_RAD:
            return math.radians(angle)
        
        elif units == ANGLE_DEG:
            return math.degrees(angle)


class ColorProperties(PropertySet):
    """
    Collection of properties defining a color.
    
    Properties:
        
        color: pero.Color, (int,), str, callable, None or UNDEF
            Specifies the color as an RGB or RGBA tuple, hex code, name or
            pero.Color. If the color is set to None, transparent color is
            set instead.
        
        alpha: int, callable, None or UNDEF
            Specifies the color alpha channel as a value between 0 and 255,
            where 0 means fully transparent and 255 fully opaque. If this value
            is set, it will overwrite the alpha channel in the final color.
    """
    
    color = ColorProperty(UNDEF, nullable=True)
    alpha = RangeProperty(None, minimum=0, maximum=255, nullable=True)
    
    
    @staticmethod
    def get_color(prop_set, prefix="", source=UNDEF, overrides=None):
        """
        Gets current color property value with the alpha property applied.
        
        Args:
            prop_set: pero.PropertySet
                Property set from which to retrieve the properties.
            
            prefix: str
                Prefix applied to color properties.
            
            source: any
                Data source to be used for retrieving the final value of
                callable properties.
            
            overrides: dict or None
                Highest priority properties to be used instead of current values.
        
        Returns:
            pero.Color or UNDEF
                Color value with applied alpha or UNDEF if color not defined.
        """
        
        # check type
        if not isinstance(prop_set, PropertySet):
            message = "Property set must be of type pero.PropertySet! -> %s" % type(prop_set)
            raise TypeError(message)
        
        # check prefix
        if prefix and prefix[-1] != PROP_SPLITTER:
            prefix = prefix + PROP_SPLITTER
        
        # get properties
        color = prop_set.get_property(prefix+'color', source, overrides)
        alpha = prop_set.get_property(prefix+'alpha', source, overrides)
        
        # make transparent from None
        if color is None:
            return Color.Transparent
        
        # check if defined
        if color is UNDEF or alpha is UNDEF or alpha is None:
            return color
        
        # apply alpha
        if color.alpha != alpha:
            color = color.opaque(alpha/255.)
        
        return color


class LineProperties(PropertySet):
    """
    Collection of properties defining a line or pen style.
    
    Properties:
        
        line_color: pero.Color, (int,), str, callable, None or UNDEF
            Specifies the line color as an RGB or RGBA tuple, hex code, name or
            pero.Color.
        
        line_alpha: int, callable, None or UNDEF
            Specifies the line color alpha channel as a value between 0 and 255,
            where 0 is fully transparent and 255 fully opaque. If this value is
            set, it will overwrite the alpha channel in the final line color.
        
        line_width: int, float, callable or UNDEF
            Specifies the line width.
        
        line_style: pero.LINE_STYLE, callable or UNDEF
            Specifies the line drawing style as any item from the
            pero.LINE_STYLE enum.
        
        line_dash: (float,), callable, None or UNDEF
            Specifies the line dash style as a collection of numbers defining
            the lengths of lines and spaces in-between. Specified value is used
            if the 'line_style' property is set to pero.LINE_STYLE_CUSTOM.
        
        line_cap: pero.LINE_CAP, callable or UNDEF
            Specifies the line ends shape as any item from the pero.LINE_CAP
            enum.
        
        line_join: pero.LINE_JOIN, callable or UNDEF
            Specifies the line corners shape as any item from the
            pero.LINE_JOIN enum.
    """
    
    color = Include(ColorProperties, "line_")
    line_width = NumProperty(1)
    line_dash = DashProperty(None, nullable=True)
    line_style = EnumProperty(LINE_STYLE_SOLID, enum=LINE_STYLE)
    line_cap = EnumProperty(LINE_CAP_ROUND, enum=LINE_CAP)
    line_join = EnumProperty(LINE_JOIN_ROUND, enum=LINE_JOIN)


class FillProperties(PropertySet):
    """
    Collection of properties defining a fill or brush style.
    
    Properties:
        
        fill_color: pero.Color, (int,), str, callable, None or UNDEF
            Specifies the fill color as an RGB or RGBA tuple, hex code, name or
            pero.Color.
        
        fill_alpha: int, callable, None or UNDEF
            Specifies the fill alpha channel as a value between 0 and 255, where
            0 is fully transparent and 255 fully opaque. If this value is set,
            it will overwrite the alpha channel in the final fill color.
        
        fill_style: pero.FILL_STYLE, callable or UNDEF
            Specifies the fill style as any item from the pero.FILL_STYLE
            enum.
    """
    
    color = Include(ColorProperties, "fill_")
    fill_style = EnumProperty(FILL_STYLE_SOLID, enum=FILL_STYLE)


class TextProperties(PropertySet):
    """
    Collection of properties defining a text style.
    
    Properties:
        
        font_size: int, callable, None or UNDEF
            Specifies the font size or None to reset to default size.
        
        font_name: str, callable, None or UNDEF
            Specifies an existing font name or None to reset to default family.
        
        font_family: pero.FONT_FAMILY, callable, None or UNDEF
            Specifies the font family as any item from the pero.FONT_FAMILY
            enum or None to reset to default family.
        
        font_style: pero.FONT_STYLE, callable, None or UNDEF
            Specifies the font style as any item from the pero.FONT_STYLE
            enum or None to reset to default style.
        
        font_weight: pero.FONT_WEIGHT, callable, None or UNDEF
            Specifies the font weight as any item from the pero.FONT_WEIGHT
            enum or None to reset to default weight.
        
        text_align: pero.TEXT_ALIGN, callable, None or UNDEF
            Specifies the text alignment as any item from the pero.TEXT_ALIGN
            enum or None to reset to default alignment.
        
        text_base: pero.TEXT_BASE, callable, None or UNDEF
            Specifies the text baseline as any item from the
            pero.TEXT_BASE enum or None to reset to default baseline.
        
        text_color: pero.Color, (int,), str, callable, None or UNDEF
            Specifies the text foreground color as an RGB or RGBA tuple, hex
            code, name or pero.Color.
        
        text_alpha: int, callable, None or UNDEF
            Specifies the text foreground alpha channel as a value between 0 and
            255, where 0 is fully transparent and 255 fully opaque. If this
            value is set, it will overwrite the alpha channel of the final text
            color.
        
        text_bgr_color: pero.Color, (int,), str, callable, None or UNDEF
            Specifies the text background color as an RGB or RGBA tuple, hex
            code, name or pero.Color.
        
        text_bgr_alpha: int, callable, None or UNDEF
            Specifies the text background alpha channel as a value between 0 and
            255, where 0 is fully transparent and 255 fully opaque. If this
            value is set, it will overwrite the alpha channel of the final text
            background color.
        
        text_split: bool
            Specifies whether the text should be first split into individual
            lines. This requires corresponding 'text_splitter' property to be
            set.
        
        text_splitter: str
            Specifies the character(s) to be used for splitting a text into
            individual lines.
        
        text_spacing: float
            Specifies additional space to be inserted between text lines as
            multiplier of line height.
    """
    
    font_size = IntProperty(11, nullable=True)
    font_name = StringProperty(UNDEF, nullable=True)
    font_family = EnumProperty(FONT_FAMILY_SANS, enum=FONT_FAMILY, nullable=True)
    font_style = EnumProperty(FONT_STYLE_NORMAL, enum=FONT_STYLE, nullable=True)
    font_weight = EnumProperty(FONT_WEIGHT_NORMAL, enum=FONT_WEIGHT, nullable=True)
    
    text_align = EnumProperty(TEXT_ALIGN_LEFT, enum=TEXT_ALIGN, nullable=True)
    text_base = EnumProperty(TEXT_BASE_TOP, enum=TEXT_BASE, nullable=True)
    
    for_color = Include(ColorProperties, prefix="text_", color='#000')
    bgr_color = Include(ColorProperties, prefix="text_bgr_", color=None)
    
    text_split = BoolProperty(True)
    text_splitter = StringProperty(LINE_SPLITTER)
    text_spacing = FloatProperty(0)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

from .. properties import *
from .. drawing import Path
from . glyph import Glyph
from . shapes import Line
from . markers import MarkerProperty, Circle


class Pather(Glyph):
    """
    Defines a utility glyph visualizing all the details of given path. This
    draws the path in a similar way as any vector graphic editor, showing all
    the anchors, control points and handles.
    
    Properties:
        
        path: pero.Path, callable, None or UNDEF
            Specifies the path to draw.
        
        show_anchors: bool, callable
            Specifies whether anchor points should be drawn.
        
        show_handles: bool, callable
            Specifies whether control points with handles should be drawn.
        
        show_cursor: bool, callable
            Specifies whether current cursor position should be drawn.
        
        anchor: pero.Marker, pero.MARKER, callable, None or UNDEF
            Specifies the marker glyph to draw the anchor points with. The value
            can be specified by any item from the pero.MARKER enum or as a
            pero.Marker instance.
        
        control: pero.Marker, pero.MARKER, callable, None or UNDEF
            Specifies the marker glyph to draw the control points with. The
            value can be specified by any item from the pero.MARKER enum or
            as a pero.Marker instance.
        
        cursor: pero.Marker, pero.MARKER, callable, None or UNDEF
            Specifies the marker glyph to draw the cursor position with. The
            value can be specified by any item from the pero.MARKER enum or
            as a pero.Marker instance.
        
        handle: pero.Line, callable, None or UNDEF
            Specifies the line glyph to draw the control handles with.
        
        line properties:
            Includes pero.LineProperties to specify the path outline.
        
        fill properties:
            Includes pero.FillProperties to specify the path fill.
    """
    
    show_anchors = BoolProperty(True)
    show_handles = BoolProperty(True)
    show_cursor = BoolProperty(True)
    
    path = Property(UNDEF, types=(Path,), nullable=True)
    
    anchor = MarkerProperty(UNDEF, nullable=True)
    control = MarkerProperty(UNDEF, nullable=True)
    cursor = MarkerProperty(UNDEF, nullable=True)
    handle = Property(UNDEF, types=(Line,), nullable=True)
    
    pen = Include(LineProperties, line_color="#000", line_width=1)
    brush = Include(FillProperties, fill_color="#ccc")
    
    
    def __init__(self, **overrides):
        """Initializes a new instance of Pather."""
        
        # init markers
        if 'anchor' not in overrides:
            overrides['anchor'] = Circle(
                size = 8,
                line_color = "#fff",
                fill_color = "#000")
        
        if 'control' not in overrides:
            overrides['control'] = Circle(
                size = 6,
                line_color = None,
                fill_color = "#77f")
        
        if 'cursor' not in overrides:
            overrides['cursor'] = Circle(
                size = 6,
                line_color = None,
                fill_color = "#f00")
        
        if 'handle' not in overrides:
            overrides['handle'] = Line(
                line_color = "#77f")
        
        # init base
        super().__init__(**overrides)
    
    
    def draw(self, canvas, source=UNDEF, **overrides):
        """Uses given canvas to draw glyph."""
        
        # check if visible
        if not self.is_visible(source, overrides):
            return
        
        # get properties
        tag = self.get_property('tag', source, overrides)
        show_anchors = self.get_property('show_anchors', source, overrides)
        show_handles = self.get_property('show_handles', source, overrides)
        show_cursor = self.get_property('show_cursor', source, overrides)
        path = self.get_property('path', source, overrides)
        anchor = self.get_property('anchor', source, overrides)
        control = self.get_property('control', source, overrides)
        cursor = self.get_property('cursor', source, overrides)
        handle = self.get_property('handle', source, overrides)
        
        # check data
        if not path:
            return
        
        # get data
        handles = path.handles()
        anchors = path.anchors()
        
        # get overrides
        anchor_overrides = self.get_child_overrides('anchor', overrides)
        control_overrides = self.get_child_overrides('control', overrides)
        cursor_overrides = self.get_child_overrides('cursor', overrides)
        handle_overrides = self.get_child_overrides('handle', overrides)
        
        # start drawing group
        canvas.group(tag, "pather")
        
        # draw path
        canvas.set_pen_by(self, source=source, overrides=overrides)
        canvas.set_brush_by(self, source=source, overrides=overrides)
        canvas.draw_path(path)
        
        # draw handles
        if show_handles and handle:
            canvas.group(None, "handles")
            for item in handles:
                handle.draw(canvas, x1=item[0], y1=item[1], x2=item[2], y2=item[3], **handle_overrides)
            canvas.ungroup()
        
        # draw anchors
        if show_anchors and anchor:
            canvas.group(None, "anchors")
            for item in anchors:
                anchor.draw(canvas, x=item[0], y=item[1], **anchor_overrides)
            canvas.ungroup()
        
        # draw controls
        if show_handles and control:
            canvas.group(None, "controls")
            for item in handles:
                control.draw(canvas, x=item[2], y=item[3], **control_overrides)
            canvas.ungroup()
        
        # draw cursor
        if show_cursor and cursor:
            cursor.draw(canvas, x=path.cursor[0], y=path.cursor[1], **cursor_overrides)
        
        # end drawing group
        canvas.ungroup()



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import math
from ..enums import *
from ..properties import *
from .formatter import Formatter


class SplitFormatter(Formatter):
    """
    This formatter tool uses defined splits to denote scale and attach
    optional 'units'. E.g. 1000 with 'Hz' units will be formatted into '1 kHz'
    label.
    
    The number of visible decimal places is determined automatically by current
    'precision' and 'domain' but it can also be specified directly using the
    'places' property.
    
    Sometimes it might be handy to just format the number and remove the whole
    suffix (the split symbol and units) from all the labels. This can be
    achieved by setting the 'hide_suffix' property to True and then the actual
    suffix can be retrieved using the 'suffix' method.
    
    Properties:
        
        prefixes: {str: int}
            Prefixes definition.
        
        units: str, None or UNDEF
            Specifies the optional units to be added right after a formatted
            label.
        
        places: int or UNDEF
            Specifies the requested number of decimal placed to show. If not
            specified, the formatting is automatically set by current
            'precision' and 'domain'.
        
        hide_suffix: bool
            Specifies whether the whole suffix (the split symbol and units)
            should be removed from the formatted values (True). The actual
            suffix can then be retrieved using the 'suffix' method.
        
        suffix_template: str, None or UNDEF
            Specifies the format()-style template to be used for suffix
            formatting.
    """
    
    splits = DictProperty(SPLITS_ENG, dynamic=False)
    units = StringProperty(UNDEF, dynamic=False, nullable=True)
    places = IntProperty(UNDEF, dynamic=False)
    
    hide_suffix = BoolProperty(False, dynamic=False)
    suffix_template = StringProperty(" ({0})", dynamic=False)
    
    
    def __init__(self, **overrides):
        """Initializes a new instance of SplitFormatter."""
        
        super().__init__(**overrides)
        
        # init buffers
        self._power = 1
        self._suffix = ""
        self._template = None
        self._is_dirty = True
        
        # bind events
        self.bind(EVT_PROPERTY_CHANGED, self._on_split_formatter_property_changed)
    
    
    def format(self, value, *args, **kwargs):
        """
        Formats given value using current formatting.

        Args:
            value: float
                Value to be formatted.

        Returns:
            str
                Formatted label.
        """
        
        # init formatting
        if self._is_dirty:
            self._init_formatting()
        
        # get template
        template = self._template
        if not template:
            template = self._make_template(value)
        
        # apply power
        if self._power:
            value /= self._power
        
        # apply template
        return template.format(value)
    
    
    def convert(self, value, *args, **kwargs):
        """
        Converts given value using current formatting.
        
        Args:
            value: float
                Value to be scaled.
        
        Returns:
            float
                Converted value.
        """
        
        # init formatting
        if self._is_dirty:
            self._init_formatting()
        
        # init template
        if not self._template:
            self._make_template(value)
        
        # apply power
        return value / self._power
    
    
    def invert(self, value, *args, **kwargs):
        """
        Inverts given value using current formatting.
        
        Args:
            value: float
                Value to be inverted.
        
        Returns:
            float
                Inverted value.
        """
        
        # init formatting
        if self._is_dirty:
            self._init_formatting()
        
        # init template
        if not self._template:
            self._make_template(value)
        
        # apply power
        return value * self._power
    
    
    def suffix(self, *args, **kwargs):
        """
        Gets current (or the latest) suffix (e.g. kHz).

        Returns:
            str
                Labels suffix.
        """
        
        # init formatting
        if self._is_dirty:
            self._init_formatting()
        
        # return suffix
        if self.hide_suffix and self._suffix:
            if self.suffix_template:
                return self.suffix_template.format(self._suffix)
            else:
                return self._suffix
        
        # return suffix
        return ""
    
    
    def _init_formatting(self):
        """Initializes formatting based on current settings."""
        
        # reset
        self._power = 0
        self._suffix = ""
        self._template = None
        self._is_dirty = False
        
        # check domain
        if not self.domain:
            return
        
        # make template
        self._template = self._make_template(abs(self.domain))
    
    
    def _make_template(self, domain):
        """Creates template to cover expected range."""
        
        # get splits
        splits = {v: k for k, v in self.splits.items()}
        splits_val = tuple(sorted(self.splits.values(), reverse=True))
        
        # get power
        self._power = 1
        for split in splits_val:
            if domain / split >= 1.:
                self._power = split
                break
        
        # crop power
        self._power = min(self._power, max(splits_val))
        self._power = max(self._power, min(splits_val))
        
        # get suffix
        self._suffix = splits.get(self._power, "")
        
        # add units
        if self.units:
            self._suffix += self.units
        
        # init direct suffix
        suffix = "" if self.hide_suffix else self._suffix
        if suffix:
            suffix = " " + suffix
        
        # get digits
        digits = 0
        
        # use specified digits
        if self.places is not UNDEF:
            digits = int(self.places)
        
        # get digits from precision
        elif self.precision:
            prec = self.precision / self._power
            prec_dig = int(math.floor(math.log10(prec)))
            digits = abs(prec_dig) if prec_dig < 0 else 0
        
        # make template
        return "{:.%df}%s" % (max(0, digits), suffix)
    
    
    def _on_split_formatter_property_changed(self, evt=None):
        """Called after a property has changed."""
        
        self._is_dirty = True


class EngFormatter(SplitFormatter):
    """Special type of pero.SplitFormatter predefined for engineering notation."""
    
    splits = DictProperty(SPLITS_ENG, dynamic=False)


class BytesFormatter(SplitFormatter):
    """Special type of pero.SplitFormatter predefined for bytes scale."""
    
    splits = DictProperty(SPLITS_BYTES, dynamic=False)
    units = StringProperty("B", dynamic=False, nullable=True)


class SecondsFormatter(SplitFormatter):
    """Special type of pero.SplitFormatter predefined for timescale."""
    
    splits = DictProperty(SPLITS_TIME, dynamic=False)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

# import main objects
from . handler import EvtHandler
from . event import Event
from . prop import PropertyChangedEvt
from . canvas import PenChangedEvt, BrushChangedEvt, TextChangedEvt
from . view import ViewEvt, SizeEvt, ZoomEvt
from . keys import KeyEvt, KeyDownEvt, KeyUpEvt
from . mouse import MouseEvt, MouseEnterEvt, MouseLeaveEvt
from . mouse import MouseMotionEvt, MouseScrollEvt
from . mouse import LeftDownEvt, LeftUpEvt, LeftDClickEvt
from . mouse import MiddleDownEvt, MiddleUpEvt, MiddleDClickEvt
from . mouse import RightDownEvt, RightUpEvt, RightDClickEvt
from . touch import Touch, TouchEvt, TouchStartEvt, TouchEndEvt, TouchMoveEvt, TouchCancelEvt, TouchDTapEvt



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import unittest
import pero
import numpy


class TestCase(unittest.TestCase):
    """Test case for linear interpolator."""
    
    
    def test_positive(self):
        """Tests whether interpolator works correctly for positive range."""
        
        interpol = pero.LinInterpol()
        
        # test inside
        self.assertEqual(interpol.normalize(3, 2, 4), 0.5)
        self.assertEqual(interpol.denormalize(0.5, 2, 4), 3)
        
        # test left
        self.assertEqual(interpol.normalize(1, 2, 4), -0.5)
        self.assertEqual(interpol.denormalize(-0.5, 2, 4), 1)
        
        # test right
        self.assertEqual(interpol.normalize(5, 2, 4), 1.5)
        self.assertEqual(interpol.denormalize(1.5, 2, 4), 5)
        
        # test zero
        self.assertEqual(interpol.normalize(0, 2, 4), -1.0)
        self.assertEqual(interpol.denormalize(-1.0, 2, 4), 0)
    
    
    def test_positive_reversed(self):
        """Tests whether interpolator works correctly for positive reversed range."""
        
        interpol = pero.LinInterpol()
        
        # test inside
        self.assertEqual(interpol.normalize(3, 4, 2), 0.5)
        self.assertEqual(interpol.denormalize(0.5, 4, 2), 3)
        
        # test left
        self.assertEqual(interpol.normalize(5, 4, 2), -0.5)
        self.assertEqual(interpol.denormalize(-0.5, 4, 2), 5)
        
        # test right
        self.assertEqual(interpol.normalize(1, 4, 2), 1.5)
        self.assertEqual(interpol.denormalize(1.5, 4, 2), 1)
        
        # test zero
        self.assertEqual(interpol.normalize(0, 4, 2), 2.0)
        self.assertEqual(interpol.denormalize(2.0, 4, 2), 0)
    
    
    def test_negative(self):
        """Tests whether interpolator works correctly for negative range."""
        
        interpol = pero.LinInterpol()
        
        # test inside
        self.assertEqual(interpol.normalize(-3, -4, -2), 0.5)
        self.assertEqual(interpol.denormalize(0.5, -4, -2), -3)
        
        # test left
        self.assertEqual(interpol.normalize(-5, -4, -2), -0.5)
        self.assertEqual(interpol.denormalize(-0.5, -4, -2), -5)
        
        # test right
        self.assertEqual(interpol.normalize(-1, -4, -2), 1.5)
        self.assertEqual(interpol.denormalize(1.5, -4, -2), -1)
        
        # test zero
        self.assertEqual(interpol.normalize(0, -4, -2), 2.0)
        self.assertEqual(interpol.denormalize(2.0, -4, -2), 0)
    
    
    def test_negative_reversed(self):
        """Tests whether interpolator works correctly for negative reversed range."""
        
        interpol = pero.LinInterpol()
        
        # test inside
        self.assertEqual(interpol.normalize(-3, -2, -4), 0.5)
        self.assertEqual(interpol.denormalize(0.5, -2, -4), -3)
        
        # test left
        self.assertEqual(interpol.normalize(-1, -2, -4), -0.5)
        self.assertEqual(interpol.denormalize(-0.5, -2, -4), -1)
        
        # test right
        self.assertEqual(interpol.normalize(-5, -2, -4), 1.5)
        self.assertEqual(interpol.denormalize(1.5, -2, -4), -5)
        
        # test zero
        self.assertEqual(interpol.normalize(0, -2, -4), -1.0)
        self.assertEqual(interpol.denormalize(-1.0, -2, -4), 0)
    
    
    def test_zero_cross(self):
        """Tests whether interpolator works correctly for cross-zero range."""
        
        interpol = pero.LinInterpol()
        
        # test inside
        self.assertEqual(interpol.normalize(2, -2, 6), 0.5)
        self.assertEqual(interpol.denormalize(0.5, -2, 6), 2)
        
        # test left
        self.assertEqual(interpol.normalize(-4, -2, 6), -0.25)
        self.assertEqual(interpol.denormalize(-0.25, -2, 6), -4)
        
        # test right
        self.assertEqual(interpol.normalize(8, -2, 6), 1.25)
        self.assertEqual(interpol.denormalize(1.25, -2, 6), 8)
        
        # test zero
        self.assertEqual(interpol.normalize(0, -2, 6), 0.25)
        self.assertEqual(interpol.denormalize(0.25, -2, 6), 0)
    
    
    def test_zero_left(self):
        """Tests whether interpolator works correctly for left-zero range."""
        
        interpol = pero.LinInterpol()
        
        # test inside
        self.assertEqual(interpol.normalize(2, 0, 4), 0.5)
        self.assertEqual(interpol.denormalize(0.5, 0, 4), 2)
        
        # test left
        self.assertEqual(interpol.normalize(-2, 0, 4), -0.5)
        self.assertEqual(interpol.denormalize(-0.5, 0, 4), -2)
        
        # test right
        self.assertEqual(interpol.normalize(6, 0, 4), 1.5)
        self.assertEqual(interpol.denormalize(1.5, 0, 4), 6)
        
        # test zero
        self.assertEqual(interpol.normalize(0, 0, 4), 0)
        self.assertEqual(interpol.denormalize(0, 0, 4), 0)
    
    
    def test_zero_right(self):
        """Tests whether interpolator works correctly for right-zero range."""
        
        interpol = pero.LinInterpol()
        
        # test inside
        self.assertEqual(interpol.normalize(-2, -4, 0), 0.5)
        self.assertEqual(interpol.denormalize(0.5, -4, 0), -2)
        
        # test left
        self.assertEqual(interpol.normalize(-6, -4, 0), -0.5)
        self.assertEqual(interpol.denormalize(-0.5, -4, 0), -6)
        
        # test right
        self.assertEqual(interpol.normalize(2, -4, 0), 1.5)
        self.assertEqual(interpol.denormalize(1.5, -4, 0), 2)
        
        # test zero
        self.assertEqual(interpol.normalize(0, -4, 0), 1.0)
        self.assertEqual(interpol.denormalize(1.0, -4, 0), 0)
    
    
    def test_arrays(self):
        """Tests whether interpolator works correctly with arrays."""
        
        interpol = pero.LinInterpol()
        
        # test positive
        data = [0, 1, 3, 5]
        model = [-1., -0.5, 0.5, 1.5]
        
        self.assertEqual(list(interpol.normalize(numpy.array(data), 2, 4)), model)
        self.assertEqual(list(interpol.denormalize(numpy.array(model), 2, 4)), data)
        
        # test positive reversed
        data = [0, 1, 3, 5]
        model = [2., 1.5, 0.5, -0.5]
        
        self.assertEqual(list(interpol.normalize(numpy.array(data), 4, 2)), model)
        self.assertEqual(list(interpol.denormalize(numpy.array(model), 4, 2)), data)
        
        # test negative
        data = [0, -1, -3, -5]
        model = [2., 1.5, 0.5, -0.5]
        
        self.assertEqual(list(interpol.normalize(numpy.array(data), -4, -2)), model)
        self.assertEqual(list(interpol.denormalize(numpy.array(model), -4, -2)), data)
        
        # test negative reversed
        data = [0, -1, -3, -5]
        model = [-1., -0.5, 0.5, 1.5]
        
        self.assertEqual(list(interpol.normalize(numpy.array(data), -2, -4)), model)
        self.assertEqual(list(interpol.denormalize(numpy.array(model), -2, -4)), data)
        
        # test zero cross
        data = [-4, 0, 2, 8]
        model = [-0.25, 0.25, 0.5, 1.25]
        
        self.assertEqual(list(interpol.normalize(numpy.array(data), -2, 6)), model)
        self.assertEqual(list(interpol.denormalize(numpy.array(model), -2, 6)), data)
        
        # test zero left
        data = [-2, 0, 2, 6]
        model = [-0.5, 0, 0.5, 1.5]
        
        self.assertEqual(list(interpol.normalize(numpy.array(data), 0, 4)), model)
        self.assertEqual(list(interpol.denormalize(numpy.array(model), 0, 4)), data)
        
        # test zero right
        data = [-6, -2, 0, 2]
        model = [-0.5, 0.5, 1.0, 1.5]
        
        self.assertEqual(list(interpol.normalize(numpy.array(data), -4, 0)), model)
        self.assertEqual(list(interpol.denormalize(numpy.array(model), -4, 0)), data)


# run test case
if __name__ == "__main__":
    unittest.main(verbosity=2)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import ui
from ... properties import *
from ... colors import Transparent, Black
from ... drawing import Canvas, Matrix, ClipState
from . enums import *


class UICanvas(Canvas):
    """Pythonista UI drawing canvas."""
    
    
    def __init__(self, **overrides):
        """
        Initializes a new instance of UICanvas.
        
        Args:
            overrides: str:any pairs
                Value overwrites for specific properties.
        """
        
        # init buffers
        self._pen = {
            'color': Transparent,
            'width': 1,
            'cap': UI_LINE_CAP[LINE_CAP_ROUND],
            'join': UI_LINE_JOIN[LINE_JOIN_MITER],
            'dash': []}
        
        self._brush = Transparent
        
        self._font = {
            'name': '<system>',
            'size': 10,
            'for_color': Black,
            'bgr_color': Transparent}
        
        self._clipping = []
        
        # init base
        super().__init__(**overrides)
        
        # init canvas
        self._update_pen()
        self._update_brush()
        self._update_text()
        
        # bind events
        self.bind(EVT_PEN_CHANGED, self._update_pen)
        self.bind(EVT_BRUSH_CHANGED, self._update_brush)
        self.bind(EVT_TEXT_CHANGED, self._update_text)
    
    
    def get_line_size(self, text):
        """
        Gets width and height of a single text line using current text settings.
        
        Args:
            text: str
                Text for which the size should be calculated.
        
        Returns:
            (float, float)
                Line width and height.
        """
        
        # check text
        if not text:
            return 0, 0
        
        # get size
        return ui.measure_string(text, font=(self._font['name'], self._font['size']))
    
    
    def draw_path(self, path):
        """
        Draws given path using current pen and brush.
        
        Args:
            path: pero.Path
                Path to be drawn.
        """
        
        # apply scaling and offset
        matrix = Matrix()
        matrix.translate(self._offset[0], self._offset[1])
        matrix.scale(self._scale, self._scale)
        path = path.transformed(matrix)
        
        # make ui path
        ui_path = self._make_native_path(path)
        
        # apply fill
        if self._brush.alpha:
            
            ui.set_color(self._brush.rgba_r)
            ui_path.eo_fill_rule = UI_FILL_RULE[path.fill_rule]
            ui_path.fill()
        
        # apply stroke
        if self._pen['color'].alpha:
            
            ui.set_color(self._pen['color'].rgba_r)
            ui_path.line_width = self._pen['width']
            ui_path.line_cap_style = self._pen['cap']
            ui_path.line_join_style = self._pen['join']
            ui_path.set_line_dash(self._pen['dash'])
            ui_path.stroke()
    
    
    def draw_text(self, text, x, y, angle=0):
        """
        Draws a text string anchored at specified point using current text
        settings.
        
        Args:
            text: str
                Text to be drawn.
            
            x: int or float
                X-coordinate of the text anchor.
            
            y: int or float
                Y-coordinate of the text anchor.
            
            angle: int or float
                Text angle in radians.
        """
        
        # get full size
        full_width, full_height = self.get_text_size(text)
        
        # split lines
        lines = [text]
        if self.text_split and self.text_splitter:
            lines = text.split(self.text_splitter)
        
        # draw text and background
        with ui.GState():
            
            # apply angle transformation
            if angle:
                
                x = self._scale * (x + self._offset[0])
                y = self._scale * (y + self._offset[1])
                
                ui.concat_ctm(ui.Transform.translation(x, y))
                ui.concat_ctm(ui.Transform.rotation(angle))
                
                x = 0
                y = 0
            
            # draw lines
            for i, line in enumerate(lines):
                
                # init offset
                x_offset = 0
                y_offset = 0
                
                # get line size
                line_width, line_height = self.get_line_size(line)
                line_width /= self._scale
                line_height /= self._scale
                
                # adjust alignment
                if self.text_align == TEXT_ALIGN_CENTER:
                    x_offset -= 0.5*line_width
                
                elif self.text_align == TEXT_ALIGN_RIGHT:
                    x_offset -= line_width
                
                # adjust baseline
                if self.text_base == TEXT_BASE_MIDDLE:
                    y_offset -= 0.5*full_height
                
                elif self.text_base == TEXT_BASE_BOTTOM:
                    y_offset -= full_height
                
                # add line offset
                y_offset += i * line_height * (1 + self.text_spacing)
                
                # apply scaling and offset
                if angle:
                    text_x = self._scale * x_offset
                    text_y = self._scale * y_offset
                else:
                    text_x = self._scale * (x + x_offset + self._offset[0])
                    text_y = self._scale * (y + y_offset + self._offset[1])
                
                # draw background
                if self._font['bgr_color'].alpha:
                    
                    bgr_width = line_width * self._scale
                    bgr_height = line_height * self._scale
                    
                    ui.set_color(self._font['bgr_color'].rgba_r)
                    ui.fill_rect(text_x, text_y, bgr_width, bgr_height)
                
                # draw text
                if self._font['for_color'].alpha:
                    
                    ui.draw_string(line,
                        rect = (text_x, text_y, 0, 0),
                        font = (self._font['name'], self._font['size']),
                        color = self._font['for_color'].rgba_r,
                        alignment = ui.ALIGN_LEFT,
                        line_break_mode = ui.LB_WORD_WRAP)
    
    
    def clip(self, path):
        """
        Sets clipping path as intersection with current one.
        
        Args:
            path: pero.Path
                Path to be used for clipping.
        
        Returns:
            pero.ClipState
                Clipping state context manager.
        """
        
        # apply scaling and offset
        matrix = Matrix()
        matrix.translate(self._offset[0], self._offset[1])
        matrix.scale(self._scale, self._scale)
        path = path.transformed(matrix)
        
        # save current canvas state
        state = ui.GState()
        state.__enter__()
        
        # make ui path
        ui_path = self._make_native_path(path)
        
        # set clipping
        ui_path.add_clip()
        
        # remember clipping state
        self._clipping.append(state)
        
        # return state
        return ClipState(self)
    
    
    def unclip(self):
        """Removes last clipping path while keeping previous if any."""
        
        # check clip
        if not self._clipping:
            return
        
        # restore state
        state = self._clipping[-1]
        state.__exit__(None, None, None)
        
        # remove from stack
        del self._clipping[-1]
    
    
    def _make_native_path(self, path):
        """Converts given path to native path."""
        
        # init path
        ui_path = ui.Path()
        
        # convert path
        for command in path.commands():
            
            # get data
            key = command[0]
            values = command[1:]
            
            # close
            if key == PATH_CLOSE:
                ui_path.close()
            
            # move to
            elif key == PATH_MOVE:
                ui_path.move_to(*values)
            
            # line to
            elif key == PATH_LINE:
                ui_path.line_to(*values)
            
            # curve to
            elif key == PATH_CURVE:
                ui_path.add_curve(values[4], values[5], *values[:-2])
        
        return ui_path
    
    
    def _update_pen(self, evt=None):
        """Updates pen with current properties."""
        
        # get property name
        prop_name = evt.name if evt is not None else None
        
        # update color
        if prop_name is None or prop_name in ('line_color', 'line_alpha'):
            color = ColorProperties.get_color(self, "line_")
            if color is not UNDEF:
                self._pen['color'] = color
        
        # update width
        if prop_name is None or prop_name in ('line_width', 'line_scale'):
            line_width = self.line_width
            if line_width is not UNDEF:
                self._pen['width'] = line_width * self.line_scale
        
        # update cap
        if prop_name is None or prop_name == 'line_cap':
            line_cap = self.line_cap
            if line_cap is not UNDEF:
                self._pen['cap'] = UI_LINE_CAP[line_cap]
        
        # update join
        if prop_name is None or prop_name == 'line_join':
            line_join = self.line_join
            if line_join is not UNDEF:
                self._pen['join'] = UI_LINE_JOIN[line_join]
        
        # update style/dash
        if prop_name is None or prop_name in ('line_dash', 'line_style', 'line_width', 'line_scale'):
            line_style = self.line_style
            line_dash = self.line_dash if self.line_dash else []
            line_width = self._pen['width']
            
            if line_style == LINE_STYLE_SOLID:
                line_dash = []
            elif line_style not in (LINE_STYLE_CUSTOM, UNDEF):
                line_dash = UI_LINE_STYLE[line_style]
            
            self._pen['dash'] = [x*line_width for x in line_dash]
    
    
    def _update_brush(self, evt=None):
        """Updates brush with current properties."""
        
        color = ColorProperties.get_color(self, "fill_")
        
        if self.fill_style == FILL_STYLE_TRANS:
            self._brush = Transparent
        
        elif color is not UNDEF:
            self._brush = color
    
    
    def _update_text(self, evt=None):
        """Updates text with current properties."""
        
        # get property name
        prop_name = evt.name if evt is not None else None
        
        # update font
        if prop_name is None or prop_name in ('font_name', 'font_family', 'font_style', 'font_weight'):
            font = self.get_font()
            self._font['name'] = font.name
        
        # update font size
        if prop_name is None or prop_name in ('font_size', 'font_scale'):
            font_size = self.font_size
            if font_size is None:
                self._font['size'] = 11 * self.font_scale
            elif font_size is not UNDEF:
                self._font['size'] = font_size * self.font_scale
        
        # update foreground color
        if prop_name is None or prop_name in ('text_color', 'text_alpha'):
            color = ColorProperties.get_color(self, "text_")
            if color is not UNDEF:
                self._font['for_color'] = color
        
        # update background color
        if prop_name is None or prop_name in ('text_bgr_color', 'text_bgr_alpha'):
            color = ColorProperties.get_color(self, "text_bgr_")
            if color is not UNDEF:
                self._font['bgr_color'] = color



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import math
from .. enums import *
from .. properties import *
from .. drawing import Path
from . glyph import Glyph


class Annulus(Glyph):
    """
    Defines a ring-like glyph.
    
    Properties:
        
        x: int, float or callable
            Specifies the x-coordinate of the center.
        
        y: int, float or callable
            Specifies the y-coordinate of the center.
        
        inner_radius: int, float or callable
            Specifies the inner radius.
        
        outer_radius: int, float or callable
            Specifies the outer radius.
        
        line properties:
            Includes pero.LineProperties to specify the glyph outline.
        
        fill properties:
            Includes pero.FillProperties to specify the glyph fill.
    """
    
    x = NumProperty(0)
    y = NumProperty(0)
    inner_radius = NumProperty(UNDEF)
    outer_radius = NumProperty(UNDEF)
    
    line = Include(LineProperties)
    fill = Include(FillProperties)
    
    
    def draw(self, canvas, source=UNDEF, **overrides):
        """Uses given canvas to draw glyph."""
        
        # check if visible
        if not self.is_visible(source, overrides):
            return
        
        # get properties
        x = self.get_property('x', source, overrides)
        y = self.get_property('y', source, overrides)
        inner_radius = self.get_property('inner_radius', source, overrides)
        outer_radius = self.get_property('outer_radius', source, overrides)
        
        # make path
        path = Path().circle(x, y, inner_radius)
        if outer_radius:
            path.circle(x, y, outer_radius)
        
        # set pen and brush
        canvas.set_pen_by(self, source=source, overrides=overrides)
        canvas.set_brush_by(self, source=source, overrides=overrides)
        
        # draw path
        canvas.draw_path(path)


class Arc(Glyph):
    """
    Defines an arc glyph.
    
    Properties:
        
        x: int, float or callable
            Specifies the x-coordinate of the center.
        
        y: int, float or callable
            Specifies the y-coordinate of the center.
        
        radius: int, float or callable
            Specifies the arc radius.
        
        clockwise: bool or callable
            Specifies the drawing direction. If set to True the arc is drawn
            clockwise, otherwise anti-clockwise.
        
        start_angle properties:
            Includes pero.AngleProperties to specify the start angle.
        
        end_angle properties:
            Includes pero.AngleProperties to specify the end angle.
        
        line properties:
            Includes pero.LineProperties to specify the glyph outline.
        
        fill properties:
            Includes pero.FillProperties to specify the glyph fill.
    """
    
    x = NumProperty(0)
    y = NumProperty(0)
    radius = NumProperty(UNDEF)
    start_angle = Include(AngleProperties, prefix="start_")
    end_angle = Include(AngleProperties, prefix="end_")
    clockwise = BoolProperty(True)
    
    line = Include(LineProperties)
    fill = Include(FillProperties)
    
    
    def draw(self, canvas, source=UNDEF, **overrides):
        """Uses given canvas to draw glyph."""
        
        # check if visible
        if not self.is_visible(source, overrides):
            return
        
        # get properties
        x = self.get_property('x', source, overrides)
        y = self.get_property('y', source, overrides)
        start_angle = AngleProperties.get_angle(self, 'start_', ANGLE_RAD, source, overrides)
        end_angle = AngleProperties.get_angle(self, 'end_', ANGLE_RAD, source, overrides)
        radius = self.get_property('radius', source, overrides)
        clockwise = self.get_property('radius', source, overrides)
        
        # set pen and brush
        canvas.set_pen_by(self, source=source, overrides=overrides)
        canvas.set_brush_by(self, source=source, overrides=overrides)
        
        # draw
        canvas.draw_arc(x, y, radius, start_angle, end_angle, clockwise)


class Bow(Glyph):
    """
    Defines an arc-like glyph specified by radius and end-point coordinates.
    One of the four existing solutions is chosen according to the 'large' and
    'clockwise' parameters.
    
    Properties:
        
        x1: int, float, callable
            Specifies the x-coordinate of the arc start.
        
        y1: int, float, callable
            Specifies the y-coordinate of the arc start.
        
        x2: int, float, callable
            Specifies the x-coordinate of the arc end.
        
        y2: int, float, callable
            Specifies the y-coordinate of the arc end.
        
        radius: int, float, callable
            Specifies the arc radius.
        
        large: bool
            Specifies which of the possible arcs will be drawn according to
            its length.
        
        clockwise: bool, callable
            Specifies which of the possible arcs will be drawn according to
            drawing direction. If set to True the clockwise arc is drawn,
            otherwise the anti-clockwise.
    """
    
    x1 = NumProperty(0)
    y1 = NumProperty(0)
    x2 = NumProperty(0)
    y2 = NumProperty(0)
    radius = NumProperty(0)
    large = BoolProperty(False)
    clockwise = BoolProperty(True)
    
    line = Include(LineProperties)
    fill = Include(FillProperties)
    
    
    def draw(self, canvas, source=UNDEF, **overrides):
        """Uses given canvas to draw glyph."""
        
        # check if visible
        if not self.is_visible(source, overrides):
            return
        
        # get properties
        x1 = self.get_property('x1', source, overrides)
        y1 = self.get_property('y1', source, overrides)
        x2 = self.get_property('x2', source, overrides)
        y2 = self.get_property('y2', source, overrides)
        radius = self.get_property('radius', source, overrides)
        large = self.get_property('large', source, overrides)
        clockwise = self.get_property('clockwise', source, overrides)
        
        # set pen and brush
        canvas.set_pen_by(self, source=source, overrides=overrides)
        canvas.set_brush_by(self, source=source, overrides=overrides)
        
        # draw
        canvas.draw_bow(x1, y1, x2, y2, radius, large, clockwise)


class Bar(Glyph):
    """
    Defines a bar glyph.
    
    Properties:
        
        left: int, float or callable
            Specifies the x-coordinate of the left edge.
        
        right: int, float or callable
            Specifies the x-coordinate of the right edge.
        
        top: int, float or callable
            Specifies the y-coordinate of the top edge.
        
        bottom: int, float or callable
            Specifies the y-coordinate of the bottom edge.
        
        radius: int, float, (int,), (float,) callable or UNDEF
            Specifies the corner radius as a single value or values for
            individual corners starting from top-left.
        
        line properties:
            Includes pero.LineProperties to specify the glyph outline.
        
        fill properties:
            Includes pero.FillProperties to specify the glyph fill.
    """
    
    left = NumProperty(0)
    right = NumProperty(0)
    top = NumProperty(0)
    bottom = NumProperty(0)
    
    radius = QuadProperty(0)
    
    line = Include(LineProperties)
    fill = Include(FillProperties)
    
    
    def draw(self, canvas, source=UNDEF, **overrides):
        """Uses given canvas to draw glyph."""
        
        # check if visible
        if not self.is_visible(source, overrides):
            return
        
        # get properties
        left = self.get_property('left', source, overrides)
        right = self.get_property('right', source, overrides)
        top = self.get_property('top', source, overrides)
        bottom = self.get_property('bottom', source, overrides)
        radius = self.get_property('radius', source, overrides)
        
        # check coords
        if right < left:
            left, right = right, left
        if bottom < top:
            top, bottom = bottom, top
        
        # set pen and brush
        canvas.set_pen_by(self, source=source, overrides=overrides)
        canvas.set_brush_by(self, source=source, overrides=overrides)
        
        # draw
        canvas.draw_rect(left, top, right-left, bottom-top, radius)


class Ellipse(Glyph):
    """
    Defines an ellipse glyph.
    
    Properties:
        
        x: int, float or callable
            Specifies the x-coordinate of the center.
        
        y: int, float or callable
            Specifies the y-coordinate of the center.
        
        width: int, float or callable
            Specifies the full width.
        
        height: int, float or callable
            Specifies the full height.
        
        line properties:
            Includes pero.LineProperties to specify the glyph outline.
        
        fill properties:
            Includes pero.FillProperties to specify the glyph fill.
    """
    
    x = NumProperty(0)
    y = NumProperty(0)
    width = NumProperty(UNDEF)
    height = NumProperty(UNDEF)
    
    line = Include(LineProperties)
    fill = Include(FillProperties)
    
    
    def draw(self, canvas, source=UNDEF, **overrides):
        """Uses given canvas to draw glyph."""
        
        # check if visible
        if not self.is_visible(source, overrides):
            return
        
        # get properties
        x = self.get_property('x', source, overrides)
        y = self.get_property('y', source, overrides)
        width = self.get_property('width', source, overrides)
        height = self.get_property('height', source, overrides)
        
        # set pen and brush
        canvas.set_pen_by(self, source=source, overrides=overrides)
        canvas.set_brush_by(self, source=source, overrides=overrides)
        
        # draw
        canvas.draw_ellipse(x, y, width, height)


class Line(Glyph):
    """
    Defines a line glyph.
    
    Properties:
        
        x1: int, float or callable
            Specifies the x-coordinate of the start.
        
        y1: int, float or callable
            Specifies the y-coordinate of the start.
        
        x2: int, float or callable
            Specifies the x-coordinate of the end.
        
        y2: int, float or callable
            Specifies the y-coordinate of the end.
        
        line properties:
            Includes pero.LineProperties to specify the line.
    """
    
    x1 = NumProperty(0)
    y1 = NumProperty(0)
    x2 = NumProperty(0)
    y2 = NumProperty(0)
    
    line = Include(LineProperties)
    
    
    def draw(self, canvas, source=UNDEF, **overrides):
        """Uses given canvas to draw glyph."""
        
        # check if visible
        if not self.is_visible(source, overrides):
            return
        
        # get properties
        x1 = self.get_property('x1', source, overrides)
        y1 = self.get_property('y1', source, overrides)
        x2 = self.get_property('x2', source, overrides)
        y2 = self.get_property('y2', source, overrides)
        
        # set pen
        canvas.set_pen_by(self, source=source, overrides=overrides)
        
        # draw
        canvas.draw_line(x1, y1, x2, y2)


class Polygon(Glyph):
    """
    Defines a poly-line glyph.
    
    Properties:
        
        points: ((int, int),),  ((float, float),), callable, None or UNDEF
            Specifies the points as a sequence of (x,y) coordinates.
        
        line properties:
            Includes pero.LineProperties to specify the glyph outline.
        
        fill properties:
            Includes pero.FillProperties to specify the glyph fill.
    """
    
    points = TupleProperty(UNDEF, nullable=True)
    
    line = Include(LineProperties)
    fill = Include(FillProperties)
    
    
    def draw(self, canvas, source=UNDEF, **overrides):
        """Uses given canvas to draw glyph."""
        
        # check if visible
        if not self.is_visible(source, overrides):
            return
        
        # get properties
        points = self.get_property('points', source, overrides)
        
        # check data
        if not points:
            return
        
        # set pen and brush
        canvas.set_pen_by(self, source=source, overrides=overrides)
        canvas.set_brush_by(self, source=source, overrides=overrides)
        
        # draw
        canvas.draw_polygon(points)


class Ray(Glyph):
    """
    Defines a ray glyph.
    
    Properties:
        
        x: int, float or callable
            Specifies the x-coordinate of the origin.
        
        y: int, float or callable
            Specifies the y-coordinate of the origin.
        
        length: int, float or callable
            Specifies the line length.
        
        offset: int, float, callable or UNDEF
            Specifies the shift from the origin while keeping the length and
            angle.
        
        angle properties:
            Includes pero.AngleProperties to specify the line angle.
        
        line properties:
            Includes pero.LineProperties to specify the glyph outline.
    """
    
    x = NumProperty(0)
    y = NumProperty(0)
    length = NumProperty(UNDEF)
    offset = NumProperty(0)
    angle = Include(AngleProperties)
    
    line = Include(LineProperties)
    
    
    def draw(self, canvas, source=UNDEF, **overrides):
        """Uses given canvas to draw glyph."""
        
        # check if visible
        if not self.is_visible(source, overrides):
            return
        
        # get properties
        x = self.get_property('x', source, overrides)
        y = self.get_property('y', source, overrides)
        length = self.get_property('length', source, overrides)
        offset = self.get_property('offset', source, overrides)
        angle = AngleProperties.get_angle(self, '', ANGLE_RAD, source, overrides)
        
        # set pen
        canvas.set_pen_by(self, source=source, overrides=overrides)
        
        # draw
        canvas.draw_ray(x, y, angle, length, offset)


class Rect(Glyph):
    """
    Defines a rectangle glyph.
    
    Properties:
        
        x: int, float or callable
            Specifies the x-coordinate of the anchor.
        
        y: int, float or callable
            Specifies the y-coordinate of the anchor.
        
        width: int, float or callable
            Specifies the full width.
        
        height: int, float or callable
            Specifies the full height.
        
        radius: int, float, (int,), (float,) callable or UNDEF
            Specifies the corner radius as a single value or values for
            individual corners starting from top-left.
        
        anchor: pero.POSITION_COMPASS or callable
            Specifies the anchor position as any item from the
            pero.POSITION_COMPASS enum.
        
        line properties:
            Includes pero.LineProperties to specify the glyph outline.
        
        fill properties:
            Includes pero.FillProperties to specify the glyph fill.
    """
    
    x = NumProperty(0)
    y = NumProperty(0)
    width = NumProperty(UNDEF)
    height = NumProperty(UNDEF)
    
    radius = QuadProperty(0)
    anchor = EnumProperty(UNDEF, enum=POSITION_COMPASS)
    
    line = Include(LineProperties)
    fill = Include(FillProperties)
    
    
    def draw(self, canvas, source=UNDEF, **overrides):
        """Uses given canvas to draw glyph."""
        
        # check if visible
        if not self.is_visible(source, overrides):
            return
        
        # get properties
        x = self.get_property('x', source, overrides)
        y = self.get_property('y', source, overrides)
        anchor = self.get_property('anchor', source, overrides)
        width = self.get_property('width', source, overrides)
        height = self.get_property('height', source, overrides)
        radius = self.get_property('radius', source, overrides)
        
        # shift anchor
        if anchor is UNDEF or anchor == POS_NW:
            pass
        elif anchor == POS_N:
            x -= 0.5 * width
        elif anchor == POS_NE:
            x -= width
        elif anchor == POS_E:
            x -= width
            y -= 0.5 * height
        elif anchor == POS_SE:
            x -= width
            y -= height
        elif anchor == POS_S:
            x -= 0.5 * width
            y -= height
        elif anchor == POS_SW:
            y -= height
        elif anchor == POS_W:
            y -= 0.5 * height
        elif anchor == POS_C:
            x -= 0.5 * width
            y -= 0.5 * height
        
        # set pen and brush
        canvas.set_pen_by(self, source=source, overrides=overrides)
        canvas.set_brush_by(self, source=source, overrides=overrides)
        
        # draw
        canvas.draw_rect(x, y, width, height, radius)


class Shape(Glyph):
    """
    Defines a path-based glyph.
    
    Properties:
        
        path: pero.Path, callable, None or UNDEF
            Specifies the path to be drawn.
        
        line properties:
            Includes pero.LineProperties to specify the glyph outline.
        
        fill properties:
            Includes pero.FillProperties to specify the glyph fill.
    """
    
    path = Property(UNDEF, types=(Path,), nullable=True)
    
    line = Include(LineProperties)
    fill = Include(FillProperties)
    
    
    def draw(self, canvas, source=UNDEF, **overrides):
        """Uses given canvas to draw glyph."""
        
        # check if visible
        if not self.is_visible(source, overrides):
            return
        
        # get properties
        path = self.get_property('path', source, overrides)
        
        # check data
        if not path:
            return
        
        # set pen and brush
        canvas.set_pen_by(self, source=source, overrides=overrides)
        canvas.set_brush_by(self, source=source, overrides=overrides)
        
        # draw
        canvas.draw_path(path)


class Wedge(Glyph):
    """
    Defines a wedge glyph.
    
    Properties:
        
        x: int, float or callable
            Specifies the x-coordinate of the center.
        
        y: int, float or callable
            Specifies the y-coordinate of the center.
        
        offset: int, float, callable or UNDEF
            Specifies the shift from the center in the direction of an angle
            between start angle and end angle.
        
        inner_radius: int, float or callable
            Specifies the inner radius.
        
        outer_radius: int, float or callable
            Specifies the outer radius.
        
        start_angle properties:
            Includes pero.AngleProperties to specify the start angle.
        
        end_angle properties:
            Includes pero.AngleProperties to specify the end angle.
        
        clockwise: bool or callable
            Specifies the drawing direction. If set to True the arc is drawn
            clockwise, otherwise anti-clockwise.
        
        line properties:
            Includes pero.LineProperties to specify the glyph outline.
        
        fill properties:
            Includes pero.FillProperties to specify the glyph fill.
    """
    
    x = NumProperty(0)
    y = NumProperty(0)
    offset = NumProperty(0)
    
    inner_radius = NumProperty(UNDEF)
    outer_radius = NumProperty(UNDEF)
    
    start_angle = Include(AngleProperties, prefix="start_")
    end_angle = Include(AngleProperties, prefix="end_")
    clockwise = BoolProperty(True)
    
    line = Include(LineProperties)
    fill = Include(FillProperties)
    
    
    def draw(self, canvas, source=UNDEF, **overrides):
        """Uses given canvas to draw glyph."""
        
        # check if visible
        if not self.is_visible(source, overrides):
            return
        
        # get properties
        x = self.get_property('x', source, overrides)
        y = self.get_property('y', source, overrides)
        start_angle = AngleProperties.get_angle(self, 'start_', ANGLE_RAD, source, overrides)
        end_angle = AngleProperties.get_angle(self, 'end_', ANGLE_RAD, source, overrides)
        inner_radius = self.get_property('inner_radius', source, overrides)
        outer_radius = self.get_property('outer_radius', source, overrides)
        clockwise = self.get_property('clockwise', source, overrides)
        offset = self.get_property('offset', source, overrides)
        
        # apply offset
        if offset:
            offset_angle = (end_angle + start_angle) * 0.5
            x += math.cos(offset_angle) * offset
            y += math.sin(offset_angle) * offset
        
        # init path
        path = Path()
        
        # skip drawing
        if start_angle == end_angle:
            pass
        
        # draw as annulus
        elif start_angle % (2 * math.pi) == end_angle % (2 * math.pi):
            path.circle(x, y, outer_radius)
            if inner_radius:
                path.circle(x, y, inner_radius)
        
        # draw as wedge
        else:
            path.arc(x, y, outer_radius, start_angle, end_angle, clockwise)
            path.line_to(x + inner_radius*math.cos(end_angle), y + inner_radius*math.sin(end_angle))
            if inner_radius:
                path.arc_around(x, y, start_angle, not clockwise)
            path.close()
        
        # set pen and brush
        canvas.set_pen_by(self, source=source, overrides=overrides)
        canvas.set_brush_by(self, source=source, overrides=overrides)
        
        # draw path
        canvas.draw_path(path)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import math
from .. properties import *
from .. formatters import ScalarFormatter
from . utils import *
from . ticker import Ticker


class LogTicker(Ticker):
    """
    This type of ticker generates nice looking ticks and labels according to
    logarithmic scale.
    
    Properties:
        
        base: int or float
            Specifies the logarithm base.
        
        major_count: int
            Specifies the expected number of major ticks to be generated for
            current range. The final number of ticks should be close but not
            necessary the same.
        
        minor_count: int
            Specifies the expected number of minor ticks to be generated between
            two major ticks. The final number of ticks should be close but not
            necessary the same.
    """
    
    base = NumProperty(10, dynamic=False)
    major_count = IntProperty(7, dynamic=False)
    minor_count = IntProperty(4, dynamic=False)
    
    
    def __init__(self, **overrides):
        """Initializes a new instance of LogTicker."""
        
        # init formatter
        if 'formatter' not in overrides:
            overrides['formatter'] = ScalarFormatter()
        
        # init base
        super().__init__(**overrides)
    
    
    def make_ticks(self):
        """
        Generates ticks according to current settings.
        
        Returns:
            (float,), (float,)
                Generated major and minor ticks.
        """
        
        # get range
        start, end = self.start, self.end
        
        # check order
        flip = False
        if start > end:
            start, end = end, start
            flip = True
        
        # make major ticks
        major_ticks, step, stage = self._make_major_ticks(start, end, self.major_count, self.base)
        
        # make minor ticks
        minor_ticks = self._make_minor_ticks(start, end, self.minor_count, self.base, step, stage)
        
        # crop ticks to current range
        major_ticks = [x for x in major_ticks if start <= x <= end]
        minor_ticks = [x for x in minor_ticks if start <= x <= end]
        
        # flip order back
        if flip:
            major_ticks.reverse()
            minor_ticks.reverse()
        
        # update formatter
        if step < self.base:
            self.formatter.precision = step
        else:
            self.formatter.precision = end
        
        return tuple(major_ticks), tuple(minor_ticks)
    
    
    def beautify(self, start, end):
        """
        Calculates visually nice range to cover given range.
        
        Args:
            start: int or float
                Minimum value of the range.
            
            end: int or float
                Maximum value of the range.
        
        Returns:
            (float, float)
                New extended range.
        """
        
        # check order
        flip = False
        if start > end:
            start, end = end, start
            flip = True
        
        # recalculate start
        start = math.log(start, self.base)
        start = math.floor(start)
        start = math.pow(self.base, start)
        
        # recalculate end
        end = math.log(end, self.base)
        end = math.ceil(end)
        end = math.pow(self.base, end)
        
        return (start, end) if not flip else (end, start)
    
    
    def _make_major_ticks(self, start, end, count, base):
        """Makes major ticks."""
        
        lo = math.log(start, base)
        hi = math.log(end, base)
        domain = abs(hi - lo)
        
        stage1 = (count-domain)
        stage2 = (count-domain*(base-1))
        
        if domain >= count or abs(stage1) <= abs(stage2):
            
            count = min(domain, count)
            splits = (base, 1) if domain <= count else (1, 2, 5)
            step = calc_step_size(domain, count, splits, base)
            
            ticks = make_lin_ticks(lo, hi, step)
            ticks = tuple(map(lambda t: math.pow(base, t), ticks))
            
            return ticks, math.pow(base, step), 1
        
        domain = abs(end - start)
        splits = (base, 1) if base != 10 else (1, 2, 5)
        step = calc_step_size(domain, count, splits, base)
        stage3 = (count - domain/step)
        
        if stage2 < stage3:
            ticks = make_log_ticks(start, end, base)
            return ticks, step, 2
        
        else:
            ticks = make_lin_ticks(start, end, step)
            return ticks, step, 3
    
    
    def _make_minor_ticks(self, start, end, count, base, step, stage):
        """Makes minor ticks."""
        
        if stage != 1:
            step = calc_step_size(step, count, (base, 1), base)
            return make_lin_ticks(start, end, step)
        
        if step <= base:
            return make_log_ticks(start, end, base)
        
        start = math.log(start, base)
        end = math.log(end, base)
        
        step = math.log(step, base)
        step = calc_step_size(step, count, (1, 2, 5), base)
        step = max(step, 1.)
        
        ticks = make_lin_ticks(start, end, step)
        ticks = tuple(map(lambda t: math.pow(base, t), ticks))
        
        return ticks



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

from .. enums import *
from . view import ViewEvt


class MouseEvt(ViewEvt):
    """
    Defines a generic event which is fired on any mouse-related event.
    
    Attributes:
        
        x_pos: int or float
            Logical x-coordinate in device units.
        
        y_pos: int or float
            Logical y-coordinate in device units.
        
        x_rot: int
            Mouse wheel rotation in x direction.
        
        y_rot: int
            Mouse wheel rotation in y direction.
        
        left_down: bool
            Indicates left mouse button state.
        
        middle_down: bool
            Indicates middle mouse button state.
        
        right_down: bool
            Indicates right mouse button state.
        
        alt_down: bool
            Indicates Alt key state.
        
        cmd_down: bool
            Indicates Command key state.
        
        ctrl_down: bool
            Indicates Control key state.
        
        shift_down: bool
            Indicates Shift key state.
    """
    
    TYPE = EVT_MOUSE
    
    
    def __init__(self, **kwargs):
        """Initializes a new instance of MouseEvt."""
        
        self.x_pos = None
        self.y_pos = None
        
        self.x_rot = None
        self.y_rot = None
        
        self.left_down = None
        self.middle_down = None
        self.right_down = None
        
        self.alt_down = None
        self.cmd_down = None
        self.ctrl_down = None
        self.shift_down = None
        
        super().__init__(**kwargs)
    
    
    @classmethod
    def from_evt(cls, evt):
        """
        Initializes a new instance of given class by copying all data.
        
        Args:
            evt: pero.MouseEvt
                Source event from which to copy the data.
        
        Returns:
            cls instance
                New instance of requested class.
        """
        
        return cls(
            
            native = evt.native,
            view = evt.view,
            control = evt.control,
            
            x_pos = evt.x_pos,
            y_pos = evt.y_pos,
            
            x_rot = evt.x_rot,
            y_rot = evt.y_rot,
            
            left_down = evt.left_down,
            middle_down = evt.middle_down,
            right_down = evt.right_down,
            
            alt_down = evt.alt_down,
            cmd_down = evt.cmd_down,
            ctrl_down = evt.ctrl_down,
            shift_down = evt.shift_down)


class MouseEnterEvt(ViewEvt):
    """Defines an event which is fired if mouse enters window."""
    
    TYPE = EVT_MOUSE_ENTER


class MouseLeaveEvt(ViewEvt):
    """Defines an event which is fired if mouse leaves window."""
    
    TYPE = EVT_MOUSE_LEAVE


class MouseMotionEvt(MouseEvt):
    """Defines an event which is fired if mouse moves."""
    
    TYPE = EVT_MOUSE_MOTION


class MouseScrollEvt(MouseEvt):
    """Defines an event which is fired if mouse wheel rotates."""
    
    TYPE = EVT_MOUSE_SCROLL


class LeftDownEvt(MouseEvt):
    """Defines an event which is fired if left-mouse button is pressed."""
    
    TYPE = EVT_LEFT_DOWN


class LeftUpEvt(MouseEvt):
    """Defines an event which is fired if left-mouse button is released."""
    
    TYPE = EVT_LEFT_UP


class LeftDClickEvt(MouseEvt):
    """Defines an event which is fired if left-mouse button is double-clicked."""
    
    TYPE = EVT_LEFT_DCLICK


class MiddleDownEvt(MouseEvt):
    """Defines an event which is fired if middle-mouse button is pressed."""
    
    TYPE = EVT_MIDDLE_DOWN


class MiddleUpEvt(MouseEvt):
    """Defines an event which is fired if middle-mouse button is released."""
    
    TYPE = EVT_MIDDLE_UP


class MiddleDClickEvt(MouseEvt):
    """Defines an event which is fired if middle-mouse button is double-clicked."""
    
    TYPE = EVT_MIDDLE_DCLICK


class RightDownEvt(MouseEvt):
    """Defines an event which is fired if right-mouse button is pressed."""
    
    TYPE = EVT_RIGHT_DOWN


class RightUpEvt(MouseEvt):
    """Defines an event which is fired if right-mouse button is released."""
    
    TYPE = EVT_RIGHT_UP


class RightDClickEvt(MouseEvt):
    """Defines an event which is fired if right-mouse button is double-clicked."""
    
    TYPE = EVT_RIGHT_DCLICK



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

from .. enums import *
from . prop import PropertyChangedEvt


class PenChangedEvt(PropertyChangedEvt):
    """
    Defines an event which is fired if any pen-related property was changed.
    """
    
    TYPE = EVT_PEN_CHANGED


class BrushChangedEvt(PropertyChangedEvt):
    """
    Defines an event which is fired if any brush-related property was changed.
    """
    
    TYPE = EVT_BRUSH_CHANGED


class TextChangedEvt(PropertyChangedEvt):
    """
    Defines an event which is fired if any text-related property was changed.
    """
    
    TYPE = EVT_TEXT_CHANGED



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import unittest
import pero


class TestCase(unittest.TestCase):
    """Test case for logarithmic ticker with base 2."""
    
    
    def test_major_ticks(self):
        """Tests whether major ticks are generated works correctly."""
        
        # init ticker
        ticker = pero.LogTicker(base=2, major_count=7)
        
        # test above one
        ticker(start=1.1, end=0.9e3)
        ticks = ticker.major_ticks()
        self.assertEqual(ticks, (2, 4, 8, 16, 32, 64, 128, 256, 512))
        
        # test one
        ticker(start=1, end=0.9e3)
        ticks = ticker.major_ticks()
        self.assertEqual(ticks, (1, 2, 4, 8, 16, 32, 64, 128, 256, 512))
        
        # test below one
        ticker(start=0.1, end=0.9)
        ticks = ticker.major_ticks()
        self.assertEqual(ticks, (0.125, 0.25, 0.5))
        
        # test cross one
        ticker(start=0.1, end=9)
        ticks = ticker.major_ticks()
        self.assertEqual(ticks, (0.125, 0.25, 0.5, 1, 2, 4, 8))
        
        # test condensed
        ticker(start=1, end=1e7)
        ticks = ticker.major_ticks()
        self.assertEqual(ticks, (1, 16, 256, 4096, 65536, 1048576))
        
        # test flipped
        ticker(start=9, end=0.1)
        ticks = ticker.major_ticks()
        self.assertEqual(ticks, (8, 4, 2, 1, 0.5, 0.25, 0.125))
    
    
    def test_minor_ticks(self):
        """Tests whether minor ticks are generated correctly."""
        
        # init ticker
        ticker = pero.LogTicker(base=2, major_count=7, minor_count=4)
        
        # test above one
        ticker(start=1.1, end=0.9e3)
        ticks = ticker.minor_ticks()
        self.assertEqual(ticks, (2, 4, 8, 16, 32, 64, 128, 256, 512))
        
        # test one
        ticker(start=1, end=0.9e3)
        ticks = ticker.minor_ticks()
        self.assertEqual(ticks, (1, 2, 4, 8, 16, 32, 64, 128, 256, 512))
        
        # test below one
        ticker(start=0.1, end=0.9)
        ticks = ticker.minor_ticks()
        self.assertEqual(ticks, (0.125, 0.25, 0.5))
        
        # test cross one
        ticker(start=0.1, end=9)
        ticks = ticker.minor_ticks()
        self.assertEqual(ticks, (0.125, 0.25, 0.5, 1, 2, 4, 8))
        
        # test condensed
        ticker(start=1, end=1e4)
        ticks = ticker.minor_ticks()
        self.assertEqual(ticks, (1, 2, 4, 8, 16, 32, 64, 128, 256, 512, 1024, 2048, 4096, 8192))
        
        # test flipped
        ticker(start=9, end=0.1)
        ticks = ticker.minor_ticks()
        self.assertEqual(ticks, (8, 4, 2, 1, 0.5, 0.25, 0.125))
    
    
    def test_beautify(self):
        """Tests whether beautify function works correctly."""
        
        # init ticker
        ticker = pero.LogTicker(base=2, major_count=7)
        
        # test above one
        beautified = ticker.beautify(1.1, 0.9e3)
        self.assertEqual(beautified, (1, 1024))
        
        # test one
        beautified = ticker.beautify(1, 0.9e3)
        self.assertEqual(beautified, (1, 1024))
        
        # test below one
        beautified = ticker.beautify(0.1, 0.9)
        self.assertEqual(beautified, (0.0625, 1))
        
        # test cross one
        beautified = ticker.beautify(0.9, 1.1e3)
        self.assertEqual(beautified, (0.5, 2048))
        
        # test flipped
        beautified = ticker.beautify(0.9e3, 0.9)
        self.assertEqual(beautified, (1024, 0.5))


# run test case
if __name__ == "__main__":
    unittest.main(verbosity=2)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

from . loader import QWidget, QEvent, QPainter, QPicture, QPixmap
from ... events import *
from .. view import View
from . enums import *
from . canvas import QtCanvas


class QtView(QWidget, View, metaclass=type('QtViewMeta', (type(QWidget), type(View)), {})):
    """Wrapper for QWidget View."""
    
    
    def __init__(self, parent=None):
        """Initializes a new instance of QtView."""
        
        # init base
        super(QtView, self).__init__(parent)
        View.__init__(self)
        self.setMouseTracking(True)
        
        # init buffers
        self._dc_buffer = None
        self._dc_overlay = None
        
        # set window events
        self.paintEvent = self._on_paint
        self.resizeEvent = self._on_size
        
        self.keyPressEvent = self._on_key_down
        self.keyReleaseEvent = self._on_key_up
        
        self.mouseMoveEvent = self._on_mouse_move
        self.wheelEvent = self._on_mouse_wheel
        
        self.enterEvent = self._on_mouse_enter
        self.leaveEvent = self._on_mouse_leave
        
        self.mousePressEvent = self._on_mouse_down
        self.mouseReleaseEvent = self._on_mouse_up
        self.mouseDoubleClickEvent = self._on_mouse_dclick
    
    
    def set_cursor(self, cursor):
        """
        Sets given mouse cursor.
        
        Args:
            cursor: pero.CURSOR
                Cursor type to be set as any item from the pero.CURSOR enum.
        """
        
        # get qt cursor
        qt_cursor = cursor
        if qt_cursor in QT_CURSORS:
            qt_cursor = QT_CURSORS[qt_cursor]
        
        # set cursor
        self.setCursor(qt_cursor)
    
    
    def set_tooltip(self, text):
        """
        Sets given text as a system tooltip.
        
        Args:
            text: str
                Tooltip text to be shown.
        """
        
        self.setToolTip(text)
    
    
    def draw_control(self):
        """Draws current control graphics."""
        
        # init buffer
        if self._dc_buffer is None:
            dpr = self.devicePixelRatioF()
            self._dc_buffer = QPixmap(int(self.width() * dpr), int(self.height() * dpr))
            self._dc_buffer.setDevicePixelRatio(dpr)
        
        # init painter
        qp = QPainter()
        qp.begin(self._dc_buffer)
        qp.setRenderHint(QPainter.RenderHint.Antialiasing)
        qp.setRenderHint(QPainter.RenderHint.SmoothPixmapTransform)
        
        # draw control
        if self.control is not None:
            canvas = QtCanvas(qp, width=self.width(), height=self.height())
            self.control.draw(canvas)
        
        # end drawing
        qp.end()
        
        # reset overlay
        self._dc_overlay = None
        
        # update screen
        self.update()
    
    
    def draw_overlay(self, func=None, view=None, **kwargs):
        """
        Draws cursor rubber band overlay.
        
        Specified function is expected to be called with a canvas as the first
        argument followed by given overrides (i.e. func(canvas, **kwargs)).
        If the 'func' parameter is set to None current overlay is cleared.
        
        Args:
            func: callable or None
                Drawing function to be called to draw the overlay. If set to
                None, current overlay will be cleared.
            
            view: (int, int, int, int) or None
                Rectangle defined as (x, y, width, height) used to shift origin
                of drawing canvas submitted to given drawing function.
            
            kwargs: str:any pairs
                Keyword arguments, which should be provided to the given drawing
                function.
        """
        
        # skip cleaning if empty already
        if func is None and self._dc_overlay is None:
            return
        
        # clear current tooltip
        self.setToolTip("")
        
        # just clean
        if func is None:
            self._dc_overlay = None
            self.update()
            return
        
        # init buffer
        if self._dc_overlay is None:
            self._dc_overlay = QPicture()
        
        # init painter
        qp = QPainter()
        qp.begin(self._dc_overlay)
        qp.setRenderHint(QPainter.RenderHint.Antialiasing)
        qp.setRenderHint(QPainter.RenderHint.SmoothPixmapTransform)
        
        # init canvas
        canvas = QtCanvas(qp, width=self.width(), height=self.height())
        
        # draw overlay
        if view is not None:
            with canvas.view(*view, relative=True):
                func(canvas, **kwargs)
        else:
            func(canvas, **kwargs)
        
        # end drawing
        qp.end()
        
        # update screen
        self.update()
    
    
    def _init_view_event(self, evt):
        """Initialize view event."""
        
        # init base event
        view_evt = ViewEvt(
            
            native = evt,
            view = self,
            control = self.control)
        
        return view_evt
    
    
    def _init_key_event(self, evt):
        """Initialize key event."""
        
        # get key
        key = evt.key()
        char = evt.text()
        
        # convert to known key
        if key in QT_KEYS:
            key = QT_KEYS[key]
        
        # init base event
        key_evt = KeyEvt(
            
            native = evt,
            view = self,
            control = self.control,
            
            key = key,
            char = char,
            
            alt_down = bool(evt.modifiers() & Qt.KeyboardModifier.AltModifier),
            cmd_down = bool(evt.modifiers() & Qt.KeyboardModifier.ControlModifier),
            ctrl_down = bool(evt.modifiers() & Qt.KeyboardModifier.ControlModifier),
            shift_down = bool(evt.modifiers() & Qt.KeyboardModifier.ShiftModifier))
        
        return key_evt
    
    
    def _init_mouse_event(self, evt):
        """Initialize mouse event."""
        
        # init base event
        mouse_evt = MouseEvt(
            
            native = evt,
            view = self,
            control = self.control,
            
            x_pos = evt.position().x(),
            y_pos = evt.position().y(),
            
            left_down = bool(evt.buttons() & Qt.MouseButton.LeftButton),
            middle_down = bool(evt.buttons() & Qt.MouseButton.MiddleButton),
            right_down = bool(evt.buttons() & Qt.MouseButton.RightButton),
            
            alt_down = bool(evt.modifiers() & Qt.KeyboardModifier.AltModifier),
            cmd_down = bool(evt.modifiers() & Qt.KeyboardModifier.ControlModifier),
            ctrl_down = bool(evt.modifiers() & Qt.KeyboardModifier.ControlModifier),
            shift_down = bool(evt.modifiers() & Qt.KeyboardModifier.ShiftModifier))
        
        return mouse_evt
    
    
    def _init_touch_event(self, evt):
        """Initializes touch event."""
        
        # get touches
        touches = []
        for point in evt.points():
            touches.append(Touch(
                id = point.id(),
                x_pos = point.position().x(),
                y_pos = point.position().y(),
                x_prev = point.lastPosition().x(),
                y_prev = point.lastPosition().y(),
                force = point.presure(),
                state = QT_TOUCH_STATE[point.state()]))
        
        # init base event
        touch_evt = TouchEvt(
            
            native = evt,
            view = self,
            control = self.control,
            
            touches = touches,
            
            alt_down = bool(evt.modifiers() & Qt.KeyboardModifier.AltModifier),
            cmd_down = bool(evt.modifiers() & Qt.KeyboardModifier.ControlModifier),
            ctrl_down = bool(evt.modifiers() & Qt.KeyboardModifier.ControlModifier),
            shift_down = bool(evt.modifiers() & Qt.KeyboardModifier.ShiftModifier))
        
        return touch_evt
    
    
    def _on_paint(self, evt):
        """Repaints current graphics."""
        
        # init painter
        qp = QPainter()
        qp.begin(self)
        qp.setRenderHint(QPainter.RenderHint.Antialiasing)
        qp.setRenderHint(QPainter.RenderHint.SmoothPixmapTransform)
        
        # draw control
        if self._dc_buffer is not None:
            qp.drawPixmap(0, 0, self._dc_buffer)
        
        # draw overlay
        if self._dc_overlay is not None:
            qp.drawPicture(0, 0, self._dc_overlay)
        
        # end drawing
        qp.end()
    
    
    def _on_size(self, evt):
        """Repaints current graphics when size has changed."""
        
        # reset buffers
        self._dc_buffer = None
        self._dc_overlay = None
        
        # get window size
        size = evt.size()
        width = max(1, size.width())
        height = max(1, size.height())
        
        # make size event
        size_evt = SizeEvt(
            
            native = evt,
            view = self,
            control = self.control,
            
            width = width,
            height = height)
        
        # fire event
        if self.control is not None:
            self.control.fire(size_evt)
    
    
    def _on_key_down(self, evt):
        """Handles key down event."""
        
        # init base event
        key_evt = self._init_key_event(evt)
        
        # make specific event type
        key_evt = KeyDownEvt.from_evt(key_evt)
        key_evt.pressed = True
        
        # fire event
        if self.control is not None:
            self.control.fire(key_evt)
    
    
    def _on_key_up(self, evt):
        """Handles key up event."""
        
        # init base event
        key_evt = self._init_key_event(evt)
        
        # make specific event type
        key_evt = KeyUpEvt.from_evt(key_evt)
        key_evt.pressed = False
        
        # fire event
        if self.control is not None:
            self.control.fire(key_evt)
    
    
    def _on_mouse_move(self, evt):
        """Handles mouse move event."""
        
        # init base event
        mouse_evt = self._init_mouse_event(evt)
        
        # make specific event type
        mouse_evt = MouseMotionEvt.from_evt(mouse_evt)
        
        # fire event
        if self.control is not None:
            self.control.fire(mouse_evt)
    
    
    def _on_mouse_wheel(self, evt):
        """Handles mouse wheel event."""
        
        # get rotation
        rotation = evt.angleDelta()
        
        # skip if no rotation
        if rotation.x() == 0 and rotation.y() == 0:
            return
        
        # init base event
        mouse_evt = self._init_mouse_event(evt)
        
        # make specific event type
        mouse_evt = MouseScrollEvt.from_evt(mouse_evt)
        mouse_evt.x_rot = rotation.x()
        mouse_evt.y_rot = rotation.y()
        
        # fire event
        if self.control is not None:
            self.control.fire(mouse_evt)
    
    
    def _on_mouse_enter(self, evt):
        """Handles mouse enter event."""
        
        # init base event
        view_evt = self._init_view_event(evt)
        
        # make specific event type
        view_evt = MouseEnterEvt.from_evt(view_evt)
        
        # fire event
        if self.control is not None:
            self.control.fire(view_evt)
    
    
    def _on_mouse_leave(self, evt):
        """Handles mouse leave event."""
        
        # init base event
        view_evt = self._init_view_event(evt)
        
        # make specific event type
        view_evt = MouseLeaveEvt.from_evt(view_evt)
        
        # fire event
        if self.control is not None:
            self.control.fire(view_evt)
    
    
    def _on_mouse_down(self, evt):
        """Handles mouse button down event."""
        
        # init base event
        mouse_evt = self._init_mouse_event(evt)
        
        # make specific event type
        if evt.button() == Qt.MouseButton.LeftButton:
            mouse_evt = LeftDownEvt.from_evt(mouse_evt)
        
        elif evt.button() == Qt.MouseButton.MiddleButton:
            mouse_evt = MiddleDownEvt.from_evt(mouse_evt)
        
        elif evt.button() == Qt.MouseButton.RightButton:
            mouse_evt = RightDownEvt.from_evt(mouse_evt)
        
        # set focus
        self.setFocus(Qt.FocusReason.MouseFocusReason)
        
        # fire event
        if self.control is not None:
            self.control.fire(mouse_evt)
    
    
    def _on_mouse_up(self, evt):
        """Handles mouse button up event."""
        
        # init base event
        mouse_evt = self._init_mouse_event(evt)
        
        # make specific event type
        if evt.button() == Qt.MouseButton.LeftButton:
            mouse_evt = LeftUpEvt.from_evt(mouse_evt)
        
        elif evt.button() == Qt.MouseButton.MiddleButton:
            mouse_evt = MiddleUpEvt.from_evt(mouse_evt)
        
        elif evt.button() == Qt.MouseButton.RightButton:
            mouse_evt = RightUpEvt.from_evt(mouse_evt)
        
        # fire event
        if self.control is not None:
            self.control.fire(mouse_evt)
    
    
    def _on_mouse_dclick(self, evt):
        """Handles mouse button double-click event."""
        
        # init base event
        mouse_evt = self._init_mouse_event(evt)
        
        # make specific event type
        if evt.button() == Qt.MouseButton.LeftButton:
            mouse_evt = LeftDClickEvt.from_evt(mouse_evt)
        
        elif evt.button() == Qt.MouseButton.MiddleButton:
            mouse_evt = MiddleDClickEvt.from_evt(mouse_evt)
        
        elif evt.button() == Qt.MouseButton.RightButton:
            mouse_evt = RightDClickEvt.from_evt(mouse_evt)
        
        # fire event
        if self.control is not None:
            self.control.fire(mouse_evt)
    
    
    def _on_touch(self, evt):
        """Handles touch events."""
        
        # init base event
        touch_evt = self._init_touch_event(evt)
        
        # make specific event type
        if evt.type() == QEvent.TouchBegin:
            touch_evt = TouchStartEvt.from_evt(touch_evt)
        
        elif evt.type() == QEvent.TouchEnd:
            touch_evt = TouchEndEvt.from_evt(touch_evt)
        
        elif evt.type() == QEvent.TouchUpdate:
            touch_evt = TouchMoveEvt.from_evt(touch_evt)
        
        elif evt.type() == QEvent.TouchCancel:
            touch_evt = TouchCancelEvt.from_evt(touch_evt)
        
        # fire event
        if self.control is not None:
            self.control.fire(touch_evt)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import numpy


def stack(arrays, axis=1):
    """Defines the missing numpy.stack function."""
    
    sl = (slice(None),) * axis + (numpy.newaxis,)
    expanded = [arr[sl] for arr in arrays]
    return numpy.concatenate(expanded, axis=axis)


def flip(array, axis=1):
    """Defines the missing numpy.flip function."""
    
    indexer = [slice(None)] * array.ndim
    indexer[axis] = slice(None, None, -1)
    return array[tuple(indexer)]


# set to numpy if not available
if not hasattr(numpy, 'stack'):
    numpy.stack = stack

if not hasattr(numpy, 'flip'):
    numpy.flip = flip
